# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test initialization and other aspects of Angle and subclasses"""

import pickle
import threading

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

import astropy.units as u
from astropy.coordinates import (
    Angle,
    IllegalHourError,
    IllegalMinuteError,
    IllegalMinuteWarning,
    IllegalSecondError,
    IllegalSecondWarning,
    Latitude,
    Longitude,
)
from astropy.utils.compat.numpycompat import NUMPY_LT_2_0


def test_create_angles():
    """
    Tests creating and accessing Angle objects
    """

    """ The "angle" is a fundamental object. The internal
    representation is stored in radians, but this is transparent to the user.
    Units *must* be specified rather than a default value be assumed. This is
    as much for self-documenting code as anything else.

    Angle objects simply represent a single angular coordinate. More specific
    angular coordinates (e.g. Longitude, Latitude) are subclasses of Angle."""

    a1 = Angle(54.12412, unit=u.degree)
    a2 = Angle("54.12412", unit=u.degree)
    a3 = Angle("54:07:26.832", unit=u.degree)
    a4 = Angle("54.12412 deg")
    a5 = Angle("54.12412 degrees")
    a6 = Angle("54.12412°")  # because we like Unicode
    a8 = Angle("54°07'26.832\"")
    a9 = Angle([54, 7, 26.832], unit=u.degree)
    assert_allclose(a9.value, [54, 7, 26.832])
    assert a9.unit is u.degree

    a10 = Angle(3.60827466667, unit=u.hour)
    a11 = Angle("3:36:29.7888000120", unit=u.hour)

    Angle(0.944644098745, unit=u.radian)

    with pytest.raises(u.UnitsError):
        Angle(54.12412)
        # raises an exception because this is ambiguous

    with pytest.raises(u.UnitsError):
        Angle(54.12412, unit=u.m)

    with pytest.raises(ValueError):
        Angle(12.34, unit="not a unit")

    a14 = Angle("03h36m29.7888000120")  # no trailing 's', but unambiguous

    a15 = Angle("5h4m3s")  # single digits, no decimal
    assert a15.unit == u.hourangle

    a16 = Angle("1 d")
    a17 = Angle("1 degree")

    assert a16.degree == 1
    assert a17.degree == 1

    a18 = Angle("54 07.4472", unit=u.degree)
    a19 = Angle("54:07.4472", unit=u.degree)
    a20 = Angle("54d07.4472m", unit=u.degree)
    a21 = Angle("3h36m", unit=u.hour)
    a22 = Angle("3.6h", unit=u.hour)
    a23 = Angle("- 3h", unit=u.hour)
    a24 = Angle("+ 3h", unit=u.hour)
    a25 = Angle(3.0, unit=u.hour**1)

    # ensure the above angles that should match do
    assert a1 == a2 == a3 == a4 == a5 == a6 == a8 == a18 == a19 == a20
    assert_allclose(a1.radian, a2.radian)
    assert_allclose(a2.degree, a3.degree)
    assert_allclose(a3.radian, a4.radian)
    assert_allclose(a4.radian, a5.radian)
    assert_allclose(a5.radian, a6.radian)

    assert_allclose(a10.degree, a11.degree)
    assert a11 == a14
    assert a21 == a22
    assert a23 == -a24
    assert a24 == a25

    # check for illegal ranges / values
    with pytest.raises(IllegalSecondError):
        Angle("12 32 99", unit=u.degree)

    with pytest.raises(IllegalMinuteError):
        Angle("12 99 23", unit=u.degree)

    with pytest.raises(IllegalSecondError):
        Angle("12 32 99", unit=u.hour)

    with pytest.raises(IllegalMinuteError):
        Angle("12 99 23", unit=u.hour)

    with pytest.raises(IllegalHourError):
        Angle("99 25 51.0", unit=u.hour)

    with pytest.raises(ValueError):
        Angle("12 25 51.0xxx", unit=u.hour)

    with pytest.raises(ValueError):
        Angle("12h34321m32.2s")

    assert a1 is not None


def test_angle_from_view():
    q = np.arange(3.0) * u.deg
    a = q.view(Angle)
    assert type(a) is Angle
    assert a.unit is q.unit
    assert np.all(a == q)

    q2 = np.arange(4) * u.m
    with pytest.raises(u.UnitTypeError):
        q2.view(Angle)


def test_angle_ops():
    """
    Tests operations on Angle objects
    """

    # Angles can be added and subtracted. Multiplication and division by a
    # scalar is also permitted. A negative operator is also valid.  All of
    # these operate in a single dimension. Attempting to multiply or divide two
    # Angle objects will return a quantity.  An exception will be raised if it
    # is attempted to store output with a non-angular unit in an Angle [#2718].

    a1 = Angle(3.60827466667, unit=u.hour)
    a2 = Angle("54:07:26.832", unit=u.degree)
    a1 + a2  # creates new Angle object
    a1 - a2
    -a1

    assert_allclose((a1 * 2).hour, 2 * 3.6082746666700003)
    assert abs((a1 / 3.123456).hour - 3.60827466667 / 3.123456) < 1e-10

    # commutativity
    assert (2 * a1).hour == (a1 * 2).hour

    a3 = Angle(a1)  # makes a *copy* of the object, but identical content as a1
    assert_allclose(a1.radian, a3.radian)
    assert a1 is not a3

    a4 = abs(-a1)
    assert a4.radian == a1.radian

    a5 = Angle(5.0, unit=u.hour)
    assert a5 > a1
    assert a5 >= a1
    assert a1 < a5
    assert a1 <= a5

    # check operations with non-angular result give Quantity.
    a6 = Angle(45.0, u.degree)
    a7 = a6 * a5
    assert type(a7) is u.Quantity

    # but those with angular result yield Angle.
    # (a9 is regression test for #5327)
    a8 = a1 + 1.0 * u.deg
    assert type(a8) is Angle
    a9 = 1.0 * u.deg + a1
    assert type(a9) is Angle

    with pytest.raises(TypeError):
        a6 *= a5

    with pytest.raises(TypeError):
        a6 *= u.m

    with pytest.raises(TypeError):
        np.sin(a6, out=a6)


def test_angle_methods():
    # Most methods tested as part of the Quantity tests.
    # A few tests here which caused problems before: #8368
    a = Angle([0.0, 2.0], "deg")
    a_mean = a.mean()
    assert type(a_mean) is Angle
    assert a_mean == 1.0 * u.degree
    a_std = a.std()
    assert type(a_std) is Angle
    assert a_std == 1.0 * u.degree
    a_var = a.var()
    assert type(a_var) is u.Quantity
    assert a_var == 1.0 * u.degree**2
    if NUMPY_LT_2_0:
        # np.ndarray.ptp() method removed in numpy 2.0.
        a_ptp = a.ptp()
        assert type(a_ptp) is Angle
        assert a_ptp == 2.0 * u.degree
    a_max = a.max()
    assert type(a_max) is Angle
    assert a_max == 2.0 * u.degree
    a_min = a.min()
    assert type(a_min) is Angle
    assert a_min == 0.0 * u.degree


def test_angle_nan_functions():
    # Most numpy functions tested as part of the Quantity tests.
    # But check that we drop to Quantity when appropriate; see
    # https://github.com/astropy/astropy/pull/17221#discussion_r1813060768
    a = Angle([0.0, 2.0, np.nan], "deg")
    a_mean = np.nanmean(a)
    assert type(a_mean) is Angle
    assert a_mean == 1.0 * u.degree
    a_std = np.nanstd(a)
    assert type(a_std) is Angle
    assert a_std == 1.0 * u.degree
    a_var = np.nanvar(a)
    assert type(a_var) is u.Quantity
    assert a_var == 1.0 * u.degree**2
    a_max = np.nanmax(a)
    assert type(a_max) is Angle
    assert a_max == 2.0 * u.degree
    a_min = np.nanmin(a)
    assert type(a_min) is Angle
    assert a_min == 0.0 * u.degree


def test_angle_convert():
    """
    Test unit conversion of Angle objects
    """
    angle = Angle("54.12412", unit=u.degree)

    assert_allclose(angle.hour, 3.60827466667)
    assert_allclose(angle.radian, 0.944644098745)
    assert_allclose(angle.degree, 54.12412)

    assert len(angle.hms) == 3
    assert isinstance(angle.hms, tuple)
    assert angle.hms[0] == 3
    assert angle.hms[1] == 36
    assert_allclose(angle.hms[2], 29.78879999999947)
    # also check that the namedtuple attribute-style access works:
    assert angle.hms.h == 3
    assert angle.hms.m == 36
    assert_allclose(angle.hms.s, 29.78879999999947)

    assert len(angle.dms) == 3
    assert isinstance(angle.dms, tuple)
    assert angle.dms[0] == 54
    assert angle.dms[1] == 7
    assert_allclose(angle.dms[2], 26.831999999992036)
    # also check that the namedtuple attribute-style access works:
    assert angle.dms.d == 54
    assert angle.dms.m == 7
    assert_allclose(angle.dms.s, 26.831999999992036)

    assert isinstance(angle.dms[0], float)
    assert isinstance(angle.hms[0], float)

    # now make sure dms and signed_dms work right for negative angles
    negangle = Angle("-54.12412", unit=u.degree)

    assert negangle.dms.d == -54
    assert negangle.dms.m == -7
    assert_allclose(negangle.dms.s, -26.831999999992036)
    assert negangle.signed_dms.sign == -1
    assert negangle.signed_dms.d == 54
    assert negangle.signed_dms.m == 7
    assert_allclose(negangle.signed_dms.s, 26.831999999992036)


def test_angle_formatting():
    """
    Tests string formatting for Angle objects
    """

    """
    The string method of Angle has this signature:
    def string(self, unit=DEGREE, decimal=False, sep=" ", precision=5,
               pad=False):

    The "decimal" parameter defaults to False since if you need to print the
    Angle as a decimal, there's no need to use the "format" method (see
    above).
    """

    angle = Angle("54.12412", unit=u.degree)

    # __str__ is the default `format`
    assert str(angle) == angle.to_string()

    res = "Angle as HMS: 3h36m29.7888s"
    assert f"Angle as HMS: {angle.to_string(unit=u.hour)}" == res

    res = "Angle as HMS: 3:36:29.7888"
    assert f"Angle as HMS: {angle.to_string(unit=u.hour, sep=':')}" == res

    res = "Angle as HMS: 3:36:29.79"
    assert f"Angle as HMS: {angle.to_string(unit=u.hour, sep=':', precision=2)}" == res

    # Note that you can provide one, two, or three separators passed as a
    # tuple or list

    res = "Angle as HMS: 3h36m29.7888s"
    assert (
        "Angle as HMS:"
        f" {angle.to_string(unit=u.hour, sep=('h', 'm', 's'), precision=4)}" == res
    )

    res = "Angle as HMS: 3-36|29.7888"
    assert (
        f"Angle as HMS: {angle.to_string(unit=u.hour, sep=['-', '|'], precision=4)}"
        == res
    )

    res = "Angle as HMS: 3-36-29.7888"
    assert f"Angle as HMS: {angle.to_string(unit=u.hour, sep='-', precision=4)}" == res

    res = "Angle as HMS: 03h36m29.7888s"
    assert f"Angle as HMS: {angle.to_string(unit=u.hour, precision=4, pad=True)}" == res

    # Same as above, in degrees

    angle = Angle("3 36 29.78880", unit=u.degree)

    res = "Angle as DMS: 3d36m29.7888s"
    assert f"Angle as DMS: {angle.to_string(unit=u.degree)}" == res

    res = "Angle as DMS: 3:36:29.7888"
    assert f"Angle as DMS: {angle.to_string(unit=u.degree, sep=':')}" == res

    res = "Angle as DMS: 3:36:29.79"
    assert (
        f"Angle as DMS: {angle.to_string(unit=u.degree, sep=':', precision=2)}" == res
    )

    # Note that you can provide one, two, or three separators passed as a
    # tuple or list

    res = "Angle as DMS: 3d36m29.7888s"
    assert (
        f"Angle as DMS: {angle.to_string(unit=u.deg, sep=('d', 'm', 's'), precision=4)}"
        == res
    )

    res = "Angle as DMS: 3-36|29.7888"
    assert (
        f"Angle as DMS: {angle.to_string(unit=u.degree, sep=['-', '|'], precision=4)}"
        == res
    )

    res = "Angle as DMS: 3-36-29.7888"
    assert (
        f"Angle as DMS: {angle.to_string(unit=u.degree, sep='-', precision=4)}" == res
    )

    res = "Angle as DMS: 03d36m29.7888s"
    assert (
        f"Angle as DMS: {angle.to_string(unit=u.degree, precision=4, pad=True)}" == res
    )

    res = "Angle as rad: 0.0629763 rad"
    assert f"Angle as rad: {angle.to_string(unit=u.radian)}" == res

    res = "Angle as rad decimal: 0.0629763"
    assert (
        f"Angle as rad decimal: {angle.to_string(unit=u.radian, decimal=True)}" == res
    )

    # check negative angles

    angle = Angle(-1.23456789, unit=u.degree)
    angle2 = Angle(-1.23456789, unit=u.hour)

    assert angle.to_string() == "-1d14m04.444404s"
    assert angle.to_string(pad=True) == "-01d14m04.444404s"
    assert angle.to_string(unit=u.hour) == "-0h04m56.2962936s"
    assert angle2.to_string(unit=u.hour, pad=True) == "-01h14m04.444404s"
    assert angle.to_string(unit=u.radian, decimal=True) == "-0.0215473"

    # We should recognize units that are equal but not identical
    assert angle.to_string(unit=u.hour**1) == "-0h04m56.2962936s"


def test_to_string_vector():
    # Regression test for the fact that vectorize doesn't work with Numpy 1.6
    assert (
        Angle([1.0 / 7.0, 1.0 / 7.0], unit="deg").to_string()[0] == "0d08m34.28571429s"
    )
    assert Angle([1.0 / 7.0], unit="deg").to_string()[0] == "0d08m34.28571429s"
    assert Angle(1.0 / 7.0, unit="deg").to_string() == "0d08m34.28571429s"


@pytest.mark.parametrize(
    "unit, sep, expected_string",
    [
        ("deg", "fromunit", "15d00m00s"),
        ("deg", "dms", "15d00m00s"),
        ("deg", "hms", "1h00m00s"),
        ("hourangle", "fromunit", "15h00m00s"),
        ("hourangle", "dms", "225d00m00s"),
        ("hourangle", "hms", "15h00m00s"),
    ],
)
def test_angle_to_string_seps(unit, sep, expected_string):
    # see https://github.com/astropy/astropy/issues/11280
    a = Angle(15, unit)
    assert a.to_string(sep=sep) == expected_string


def test_angle_format_roundtripping():
    """
    Ensures that the string representation of an angle can be used to create a
    new valid Angle.
    """

    a1 = Angle(0, unit=u.radian)
    a2 = Angle(10, unit=u.degree)
    a3 = Angle(0.543, unit=u.degree)
    a4 = Angle("1d2m3.4s")

    assert Angle(str(a1)).degree == a1.degree
    assert Angle(str(a2)).degree == a2.degree
    assert Angle(str(a3)).degree == a3.degree
    assert Angle(str(a4)).degree == a4.degree

    # also check Longitude/Latitude
    ra = Longitude("1h2m3.4s")
    dec = Latitude("1d2m3.4s")

    assert_allclose(Angle(str(ra)).degree, ra.degree)
    assert_allclose(Angle(str(dec)).degree, dec.degree)


def test_radec():
    """
    Tests creation/operations of Longitude and Latitude objects
    """

    """
    Longitude and Latitude are objects that are subclassed from Angle. As with Angle, Longitude
    and Latitude can parse any unambiguous format (tuples, formatted strings, etc.).

    The intention is not to create an Angle subclass for every possible
    coordinate object (e.g. galactic l, galactic b). However, equatorial Longitude/Latitude
    are so prevalent in astronomy that it's worth creating ones for these
    units. They will be noted as "special" in the docs and use of the just the
    Angle class is to be used for other coordinate systems.
    """

    with pytest.raises(u.UnitsError):
        Longitude("4:08:15.162342")  # error - hours or degrees?
    with pytest.raises(u.UnitsError):
        Longitude("-4:08:15.162342")

    # the "smart" initializer allows >24 to automatically do degrees, but the
    # Angle-based one does not
    # TODO: adjust in 0.3 for whatever behavior is decided on

    # ra = Longitude("26:34:15.345634")  # unambiguous b/c hours don't go past 24
    # assert_allclose(ra.degree, 26.570929342)
    with pytest.raises(u.UnitsError):
        Longitude("26:34:15.345634")

    # ra = Longitude(68)
    with pytest.raises(u.UnitsError):
        Longitude(68)

    with pytest.raises(u.UnitsError):
        Longitude(12)

    with pytest.raises(ValueError):
        Longitude("garbage containing a d and no units")

    ra = Longitude("12h43m23s")
    assert_allclose(ra.hour, 12.7230555556)

    # Units can be specified
    ra = Longitude("4:08:15.162342", unit=u.hour)

    # TODO: this was the "smart" initializer behavior - adjust in 0.3 appropriately
    # Where Longitude values are commonly found in hours or degrees, declination is
    # nearly always specified in degrees, so this is the default.
    # dec = Latitude("-41:08:15.162342")
    with pytest.raises(u.UnitsError):
        Latitude("-41:08:15.162342")
    dec = Latitude("-41:08:15.162342", unit=u.degree)  # same as above


def test_negative_zero_dms():
    # Test for DMS parser
    a = Angle("-00:00:10", u.deg)
    assert_allclose(a.degree, -10.0 / 3600.0)

    # Unicode minus
    a = Angle("−00:00:10", u.deg)
    assert_allclose(a.degree, -10.0 / 3600.0)


def test_negative_zero_dm():
    # Test for DM parser
    a = Angle("-00:10", u.deg)
    assert_allclose(a.degree, -10.0 / 60.0)


def test_negative_zero_hms():
    # Test for HMS parser
    a = Angle("-00:00:10", u.hour)
    assert_allclose(a.hour, -10.0 / 3600.0)


def test_negative_zero_hm():
    # Test for HM parser
    a = Angle("-00:10", u.hour)
    assert_allclose(a.hour, -10.0 / 60.0)


def test_negative_sixty_hm():
    # Test for HM parser
    with pytest.warns(IllegalMinuteWarning):
        a = Angle("-00:60", u.hour)
    assert_allclose(a.hour, -1.0)


def test_plus_sixty_hm():
    # Test for HM parser
    with pytest.warns(IllegalMinuteWarning):
        a = Angle("00:60", u.hour)
    assert_allclose(a.hour, 1.0)


def test_negative_fifty_nine_sixty_dms():
    # Test for DMS parser
    with pytest.warns(IllegalSecondWarning):
        a = Angle("-00:59:60", u.deg)
    assert_allclose(a.degree, -1.0)


def test_plus_fifty_nine_sixty_dms():
    # Test for DMS parser
    with pytest.warns(IllegalSecondWarning):
        a = Angle("+00:59:60", u.deg)
    assert_allclose(a.degree, 1.0)


def test_negative_sixty_dms():
    # Test for DMS parser
    with pytest.warns(IllegalSecondWarning):
        a = Angle("-00:00:60", u.deg)
    assert_allclose(a.degree, -1.0 / 60.0)


def test_plus_sixty_dms():
    # Test for DMS parser
    with pytest.warns(IllegalSecondWarning):
        a = Angle("+00:00:60", u.deg)
    assert_allclose(a.degree, 1.0 / 60.0)


def test_angle_to_is_angle():
    with pytest.warns(IllegalSecondWarning):
        a = Angle("00:00:60", u.deg)
    assert isinstance(a, Angle)
    assert isinstance(a.to(u.rad), Angle)


def test_angle_to_quantity():
    with pytest.warns(IllegalSecondWarning):
        a = Angle("00:00:60", u.deg)
    q = u.Quantity(a)
    assert isinstance(q, u.Quantity)
    assert q.unit is u.deg


def test_quantity_to_angle():
    a = Angle(1.0 * u.deg)
    assert isinstance(a, Angle)
    with pytest.raises(u.UnitsError):
        Angle(1.0 * u.meter)
    a = Angle(1.0 * u.hour)
    assert isinstance(a, Angle)
    assert a.unit is u.hourangle
    with pytest.raises(u.UnitsError):
        Angle(1.0 * u.min)


def test_angle_string():
    with pytest.warns(IllegalSecondWarning):
        a = Angle("00:00:60", u.deg)
    assert str(a) == "0d01m00s"
    a = Angle("00:00:59S", u.deg)
    assert str(a) == "-0d00m59s"
    a = Angle("00:00:59N", u.deg)
    assert str(a) == "0d00m59s"
    a = Angle("00:00:59E", u.deg)
    assert str(a) == "0d00m59s"
    a = Angle("00:00:59W", u.deg)
    assert str(a) == "-0d00m59s"
    a = Angle("-00:00:10", u.hour)
    assert str(a) == "-0h00m10s"
    a = Angle("00:00:59E", u.hour)
    assert str(a) == "0h00m59s"
    a = Angle("00:00:59W", u.hour)
    assert str(a) == "-0h00m59s"
    a = Angle(3.2, u.radian)
    assert str(a) == "3.2 rad"
    a = Angle(4.2, u.microarcsecond)
    assert str(a) == "4.2 uarcsec"
    a = Angle("1.0uarcsec")
    assert a.value == 1.0
    assert a.unit == u.microarcsecond
    a = Angle("1.0uarcsecN")
    assert a.value == 1.0
    assert a.unit == u.microarcsecond
    a = Angle("1.0uarcsecS")
    assert a.value == -1.0
    assert a.unit == u.microarcsecond
    a = Angle("1.0uarcsecE")
    assert a.value == 1.0
    assert a.unit == u.microarcsecond
    a = Angle("1.0uarcsecW")
    assert a.value == -1.0
    assert a.unit == u.microarcsecond
    a = Angle("3d")
    assert_allclose(a.value, 3.0)
    assert a.unit == u.degree
    a = Angle("3dN")
    assert str(a) == "3d00m00s"
    assert a.unit == u.degree
    a = Angle("3dS")
    assert str(a) == "-3d00m00s"
    assert a.unit == u.degree
    a = Angle("3dE")
    assert str(a) == "3d00m00s"
    assert a.unit == u.degree
    a = Angle("3dW")
    assert str(a) == "-3d00m00s"
    assert a.unit == u.degree
    a = Angle('10"')
    assert_allclose(a.value, 10.0)
    assert a.unit == u.arcsecond
    a = Angle("10'N")
    assert_allclose(a.value, 10.0)
    assert a.unit == u.arcminute
    a = Angle("10'S")
    assert_allclose(a.value, -10.0)
    assert a.unit == u.arcminute
    a = Angle("10'E")
    assert_allclose(a.value, 10.0)
    assert a.unit == u.arcminute
    a = Angle("10'W")
    assert_allclose(a.value, -10.0)
    assert a.unit == u.arcminute
    a = Angle("45°55′12″N")
    assert str(a) == "45d55m12s"
    assert_allclose(a.value, 45.92)
    assert a.unit == u.deg
    a = Angle("45°55′12″S")
    assert str(a) == "-45d55m12s"
    assert_allclose(a.value, -45.92)
    assert a.unit == u.deg
    a = Angle("45°55′12″E")
    assert str(a) == "45d55m12s"
    assert_allclose(a.value, 45.92)
    assert a.unit == u.deg
    a = Angle("45°55′12″W")
    assert str(a) == "-45d55m12s"
    assert_allclose(a.value, -45.92)
    assert a.unit == u.deg
    with pytest.raises(ValueError):
        Angle("00h00m10sN")
    with pytest.raises(ValueError):
        Angle("45°55′12″NS")


def test_angle_repr():
    assert "Angle" in repr(Angle(0, u.deg))
    assert "Longitude" in repr(Longitude(0, u.deg))
    assert "Latitude" in repr(Latitude(0, u.deg))

    a = Angle(0, u.deg)
    repr(a)


def test_large_angle_representation():
    """Test that angles above 360 degrees can be output as strings,
    in repr, str, and to_string.  (regression test for #1413)"""
    a = Angle(350, u.deg) + Angle(350, u.deg)
    a.to_string()
    a.to_string(u.hourangle)
    repr(a)
    repr(a.to(u.hourangle))
    str(a)
    str(a.to(u.hourangle))


def test_wrap_at_inplace():
    a = Angle([-20, 150, 350, 360] * u.deg)
    out = a.wrap_at("180d", inplace=True)
    assert out is None
    assert np.all(a.degree == np.array([-20.0, 150.0, -10.0, 0.0]))


def test_latitude():
    """Test input validation for setting Latitude angles."""

    lim_exc = r"Latitude angle\(s\) must be within -90 deg <= angle <= 90 deg, got"
    with pytest.raises(ValueError, match=rf"{lim_exc} \[91. 89.\] deg"):
        Latitude([91, 89] * u.deg)
    with pytest.raises(ValueError, match=f"{lim_exc} -91.0 deg"):
        Latitude("-91d")

    lat = Latitude(["90d", "89d"])
    # check that one can get items
    assert lat[0] == 90 * u.deg
    assert lat[1] == 89 * u.deg
    # and that comparison with angles works
    assert np.all(lat == Angle(["90d", "89d"]))
    # check setitem works
    lat[1] = 45.0 * u.deg
    assert np.all(lat == Angle(["90d", "45d"]))
    # but not with values out of range
    with pytest.raises(ValueError, match=f"{lim_exc} 90.001 deg"):
        lat[0] = 90.001 * u.deg
    with pytest.raises(ValueError, match=f"{lim_exc} -90.001 deg"):
        lat[0] = -90.001 * u.deg
    # these should also not destroy input (#1851)
    assert np.all(lat == Angle(["90d", "45d"]))

    # Check error message for long-ish input arrays (#13994).
    with pytest.raises(
        ValueError, match=rf"{lim_exc} -143.239\d* deg <= angle <= 171.887\d* deg"
    ):
        lat = Latitude([0, 1, 2, 3, -2.5, -1, -0.5] * u.radian)


def test_latitude_manipulation():
    """Test value manipulation operations on Latitude angles."""

    lat = Latitude(["90d", "-29d"])
    # conserve type on unit change (closes #1423)
    angle = lat.to("radian")
    assert type(angle) is Latitude
    # but not on calculations
    angle = lat - 190 * u.deg
    assert type(angle) is Angle
    assert_allclose(angle, [-100, -219] * u.deg)

    lat = Latitude("80d")
    angle = lat / 2.0
    assert type(angle) is Angle
    assert angle == 40 * u.deg

    angle = lat * 2.0
    assert type(angle) is Angle
    assert angle == 160 * u.deg

    angle = -lat
    assert type(angle) is Angle
    assert angle == -80 * u.deg


def test_lon_as_lat():
    """Test validation when trying to interoperate with longitudes."""

    lon = Longitude(10, "deg")
    with pytest.raises(
        TypeError, match="A Latitude angle cannot be created from a Longitude angle"
    ):
        Latitude(lon)

    with pytest.raises(
        TypeError, match="A Longitude angle cannot be assigned to a Latitude angle"
    ):
        lat = Latitude([20], "deg")
        lat[0] = lon

    # Check we can work around the Lat vs Long checks by casting explicitly to Angle.
    lat = Latitude(Angle(lon))
    assert lat.value == 10.0

    # Check setitem.
    lat = Latitude([20], "deg")
    lat[0] = Angle(lon)
    assert lat.value[0] == 10.0


@pytest.mark.parametrize("lon", ["12.3dW", "12h13m12sE", ["1d", "1dW"]])
def test_lon_as_lat_str(lon):
    with pytest.raises(TypeError, match="Latitude.*cannot be created from a Longitude"):
        Latitude(lon)


@pytest.mark.parametrize("lat", ["12.3dN", "12d13m12sS", ["1d", "1dS"]])
def test_lat_as_lon_str(lat):
    with pytest.raises(TypeError, match="Longitude.*cannot be created from a Latitude"):
        Longitude(lat)


def test_longitude():
    """Test setting and manipulation operations on Longitude angles."""

    # Default wrapping at 360d with an array input
    lon = Longitude(["370d", "88d"])
    assert np.all(lon == Longitude(["10d", "88d"]))
    assert np.all(lon == Angle(["10d", "88d"]))

    # conserve type on unit change and keep wrap_angle (closes #1423)
    angle = lon.to("hourangle")
    assert type(angle) is Longitude
    assert angle.wrap_angle == lon.wrap_angle
    angle = lon[0]
    assert type(angle) is Longitude
    assert angle.wrap_angle == lon.wrap_angle
    angle = lon[1:]
    assert type(angle) is Longitude
    assert angle.wrap_angle == lon.wrap_angle

    # but not on calculations
    angle = lon / 2.0
    assert np.all(angle == Angle(["5d", "44d"]))
    assert type(angle) is Angle
    assert not hasattr(angle, "wrap_angle")

    angle = lon * 2.0 + 400 * u.deg
    assert np.all(angle == Angle(["420d", "576d"]))
    assert type(angle) is Angle

    # Test setting a mutable value and having it wrap
    lon[1] = -10 * u.deg
    assert np.all(lon == Angle(["10d", "350d"]))

    # Test wrapping and try hitting some edge cases
    lon = Longitude(np.array([0, 0.5, 1.0, 1.5, 2.0]) * np.pi, unit=u.radian)
    assert np.all(lon.degree == np.array([0.0, 90, 180, 270, 0]))

    lon = Longitude(
        np.array([0, 0.5, 1.0, 1.5, 2.0]) * np.pi, unit=u.radian, wrap_angle="180d"
    )
    assert np.all(lon.degree == np.array([0.0, 90, -180, -90, 0]))

    # Wrap on setting wrap_angle property (also test auto-conversion of wrap_angle to an Angle)
    lon = Longitude(np.array([0, 0.5, 1.0, 1.5, 2.0]) * np.pi, unit=u.radian)
    lon.wrap_angle = "180d"
    assert np.all(lon.degree == np.array([0.0, 90, -180, -90, 0]))

    lon = Longitude("460d")
    assert lon == Angle("100d")
    lon.wrap_angle = "90d"
    assert lon == Angle("-260d")

    # check that if we initialize a longitude with another longitude,
    # wrap_angle is kept by default
    lon2 = Longitude(lon)
    assert lon2.wrap_angle == lon.wrap_angle
    # but not if we explicitly set it
    lon3 = Longitude(lon, wrap_angle="180d")
    assert lon3.wrap_angle == 180 * u.deg

    # check that wrap_angle is always an Angle
    lon = Longitude(lon, wrap_angle=Longitude(180 * u.deg))
    assert lon.wrap_angle == 180 * u.deg
    assert lon.wrap_angle.__class__ is Angle

    # check that wrap_angle is not copied
    wrap_angle = 180 * u.deg
    lon = Longitude(lon, wrap_angle=wrap_angle)
    assert lon.wrap_angle == 180 * u.deg
    assert np.may_share_memory(lon.wrap_angle, wrap_angle)

    # check for problem reported in #2037 about Longitude initializing to -0
    lon = Longitude(0, u.deg)
    lonstr = lon.to_string()
    assert not lonstr.startswith("-")

    # also make sure dtype is correctly conserved
    assert Longitude(0, u.deg, dtype=float).dtype == np.dtype(float)
    assert Longitude(0, u.deg, dtype=int).dtype == np.dtype(int)


def test_lat_as_lon():
    """Test validation when trying to interoperate with latitudes."""

    lat = Latitude(10, "deg")
    # Test errors when trying to interoperate with latitudes.
    with pytest.raises(
        TypeError, match="A Longitude angle cannot be created from a Latitude angle"
    ):
        Longitude(lat)

    with pytest.raises(
        TypeError, match="A Latitude angle cannot be assigned to a Longitude angle"
    ):
        lon = Longitude([20], "deg")
        lon[0] = lat

    # Check we can work around the Lat vs Long checks by casting explicitly to Angle.
    lon = Longitude(Angle(lat))
    assert lon.value == 10.0

    # Check setitem.
    lon = Longitude([20], "deg")
    lon[0] = Angle(lat)
    assert lon.value[0] == 10.0


def test_wrap_at():
    a = Angle([-20, 150, 350, 360] * u.deg)
    assert np.all(a.wrap_at(360 * u.deg).degree == np.array([340.0, 150.0, 350.0, 0.0]))
    assert np.all(
        a.wrap_at(Angle(360, unit=u.deg)).degree == np.array([340.0, 150.0, 350.0, 0.0])
    )
    assert np.all(a.wrap_at("360d").degree == np.array([340.0, 150.0, 350.0, 0.0]))
    assert np.all(a.wrap_at("180d").degree == np.array([-20.0, 150.0, -10.0, 0.0]))
    assert np.all(
        a.wrap_at(np.pi * u.rad).degree == np.array([-20.0, 150.0, -10.0, 0.0])
    )

    # Test wrapping a scalar Angle
    a = Angle("190d")
    assert a.wrap_at("180d") == Angle("-170d")

    a = Angle(np.arange(-1000.0, 1000.0, 0.125), unit=u.deg)
    for wrap_angle in (270, 0.2, 0.0, 360.0, 500, -2000.125):
        aw = a.wrap_at(wrap_angle * u.deg)
        assert np.all(aw.degree >= wrap_angle - 360.0)
        assert np.all(aw.degree < wrap_angle)

        aw = a.to(u.rad).wrap_at(wrap_angle * u.deg)
        assert np.all(aw.degree >= wrap_angle - 360.0)
        assert np.all(aw.degree < wrap_angle)


def test_is_within_bounds():
    a = Angle([-20, 150, 350] * u.deg)
    assert a.is_within_bounds("0d", "360d") is False
    assert a.is_within_bounds(None, "360d") is True
    assert a.is_within_bounds(-30 * u.deg, None) is True

    a = Angle("-20d")
    assert a.is_within_bounds("0d", "360d") is False
    assert a.is_within_bounds(None, "360d") is True
    assert a.is_within_bounds(-30 * u.deg, None) is True


def test_angle_mismatched_unit():
    a = Angle("+6h7m8s", unit=u.degree)
    assert_allclose(a.value, 91.78333333333332)


def test_regression_formatting_negative():
    # Regression test for a bug that caused:
    #
    # >>> Angle(-1., unit='deg').to_string()
    # '-1d00m-0s'
    assert Angle(-0.0, unit="deg").to_string() == "-0d00m00s"
    assert Angle(-1.0, unit="deg").to_string() == "-1d00m00s"
    assert Angle(-0.0, unit="hour").to_string() == "-0h00m00s"
    assert Angle(-1.0, unit="hour").to_string() == "-1h00m00s"


def test_regression_formatting_default_precision():
    # Regression test for issue #11140
    assert Angle("10:20:30.12345678d").to_string() == "10d20m30.12345678s"
    assert Angle("10d20m30.123456784564s").to_string() == "10d20m30.12345678s"
    assert Angle("10d20m30.123s").to_string() == "10d20m30.123s"


def test_empty_sep():
    a = Angle("05h04m31.93830s")

    assert a.to_string(sep="", precision=2, pad=True) == "050431.94"


@pytest.mark.parametrize("angle_class", [Angle, Longitude])
@pytest.mark.parametrize("unit", [u.hourangle, u.hour, None])
def test_create_tuple_fail(angle_class, unit):
    """Creating an angle from an (h,m,s) tuple should fail."""
    with pytest.raises(TypeError, match="no longer supported"):
        angle_class((12, 14, 52), unit=unit)


def test_list_of_quantities():
    a1 = Angle([1 * u.deg, 1 * u.hourangle])
    assert a1.unit == u.deg
    assert_allclose(a1.value, [1, 15])

    a2 = Angle([1 * u.hourangle, 1 * u.deg], u.deg)
    assert a2.unit == u.deg
    assert_allclose(a2.value, [15, 1])


def test_multiply_divide():
    # Issue #2273
    a1 = Angle([1, 2, 3], u.deg)
    a2 = Angle([4, 5, 6], u.deg)
    a3 = a1 * a2
    assert_allclose(a3.value, [4, 10, 18])
    assert a3.unit == (u.deg * u.deg)

    a3 = a1 / a2
    assert_allclose(a3.value, [0.25, 0.4, 0.5])
    assert a3.unit == u.dimensionless_unscaled


def test_mixed_string_and_quantity():
    a1 = Angle(["1d", 1.0 * u.deg])
    assert_array_equal(a1.value, [1.0, 1.0])
    assert a1.unit == u.deg

    a2 = Angle(["1d", 1 * u.rad * np.pi, "3d"])
    assert_array_equal(a2.value, [1.0, 180.0, 3.0])
    assert a2.unit == u.deg


def test_array_angle_tostring():
    aobj = Angle([1, 2], u.deg)
    assert aobj.to_string().dtype.kind == "U"
    assert np.all(aobj.to_string() == ["1d00m00s", "2d00m00s"])


def test_wrap_at_without_new():
    """
    Regression test for subtle bugs from situations where an Angle is
    created via numpy channels that don't do the standard __new__ but instead
    depend on array_finalize to set state.  Longitude is used because the
    bug was in its _wrap_angle not getting initialized correctly
    """
    l1 = Longitude([1] * u.deg)
    l2 = Longitude([2] * u.deg)

    l = np.concatenate([l1, l2])
    assert l._wrap_angle is not None


def test__str__():
    """
    Check the __str__ method used in printing the Angle
    """

    # scalar angle
    scangle = Angle("10.2345d")
    strscangle = scangle.__str__()
    assert strscangle == "10d14m04.2s"

    # non-scalar array angles
    arrangle = Angle(["10.2345d", "-20d"])
    strarrangle = arrangle.__str__()

    assert strarrangle == "[10d14m04.2s -20d00m00s]"

    # summarizing for large arrays, ... should appear
    bigarrangle = Angle(np.ones(10000), u.deg)
    assert "..." in bigarrangle.__str__()


def test_repr_latex():
    """
    Check the _repr_latex_ method, used primarily by IPython notebooks
    """

    # try with both scalar
    scangle = Angle(2.1, u.deg)
    rlscangle = scangle._repr_latex_()

    # and array angles
    arrangle = Angle([1, 2.1], u.deg)
    rlarrangle = arrangle._repr_latex_()

    assert rlscangle == r"$2^\circ06{}^\prime00{}^{\prime\prime}$"
    assert rlscangle.split("$")[1] in rlarrangle

    # make sure the ... appears for large arrays
    bigarrangle = Angle(np.ones(50000) / 50000.0, u.deg)
    assert "..." in bigarrangle._repr_latex_()


def test_angle_with_cds_units_enabled():
    """Regression test for #5350

    Especially the example in
    https://github.com/astropy/astropy/issues/5350#issuecomment-248770151
    """
    # the problem is with the parser, so remove it temporarily
    from astropy.coordinates.angles.formats import _AngleParser
    from astropy.units import cds

    del _AngleParser._thread_local._parser
    with cds.enable():
        Angle("5d")
    del _AngleParser._thread_local._parser
    Angle("5d")


def test_longitude_nan():
    # Check that passing a NaN to Longitude doesn't raise a warning
    Longitude([0, np.nan, 1] * u.deg)


def test_latitude_nan():
    # Check that passing a NaN to Latitude doesn't raise a warning
    Latitude([0, np.nan, 1] * u.deg)


def test_angle_wrap_at_nan():
    # Check that no attempt is made to wrap a NaN angle
    angle = Angle([0, np.nan, 1] * u.deg)
    angle.flags.writeable = False  # to force an error if a write is attempted
    angle.wrap_at(180 * u.deg, inplace=True)


def test_angle_multithreading():
    """
    Regression test for issue #7168
    """
    angles = ["00:00:00"] * 10000

    def parse_test(i=0):
        Angle(angles, unit="hour")

    for i in range(10):
        threading.Thread(target=parse_test, args=(i,)).start()


@pytest.mark.parametrize("cls", [Angle, Longitude, Latitude])
@pytest.mark.parametrize(
    "input, expstr, exprepr",
    [
        (np.nan * u.deg, "nan", "nan deg"),
        ([np.nan, 5, 0] * u.deg, "[nan 5d00m00s 0d00m00s]", "[nan, 5., 0.] deg"),
        ([6, np.nan, 0] * u.deg, "[6d00m00s nan 0d00m00s]", "[6., nan, 0.] deg"),
        ([np.nan, np.nan, np.nan] * u.deg, "[nan nan nan]", "[nan, nan, nan] deg"),
        (np.nan * u.hour, "nan", "nan hourangle"),
        ([np.nan, 5, 0] * u.hour, "[nan 5h00m00s 0h00m00s]", "[nan, 5., 0.] hourangle"),
        ([6, np.nan, 0] * u.hour, "[6h00m00s nan 0h00m00s]", "[6., nan, 0.] hourangle"),
        (
            [np.nan, np.nan, np.nan] * u.hour,
            "[nan nan nan]",
            "[nan, nan, nan] hourangle",
        ),
        (np.nan * u.rad, "nan", "nan rad"),
        ([np.nan, 1, 0] * u.rad, "[nan 1 rad 0 rad]", "[nan, 1., 0.] rad"),
        ([1.50, np.nan, 0] * u.rad, "[1.5 rad nan 0 rad]", "[1.5, nan, 0.] rad"),
        ([np.nan, np.nan, np.nan] * u.rad, "[nan nan nan]", "[nan, nan, nan] rad"),
    ],
)
def test_str_repr_angles_nan(cls, input, expstr, exprepr):
    """
    Regression test for issue #11473
    """
    q = cls(input)
    assert str(q) == expstr
    # Deleting whitespaces since repr appears to be adding them for some values
    # making the test fail.
    assert repr(q).replace(" ", "") == f"<{cls.__name__}{exprepr}>".replace(" ", "")


@pytest.mark.parametrize("sign", (-1, 1))
@pytest.mark.parametrize(
    "value,expected_value,dtype,expected_dtype",
    [
        (np.pi * 2, 0.0, None, np.float64),
        (np.pi * 2, 0.0, np.float64, np.float64),
        (np.float32(2 * np.pi), np.float32(0.0), None, np.float32),
        (np.float32(2 * np.pi), np.float32(0.0), np.float32, np.float32),
    ],
)
def test_longitude_wrap(value, expected_value, dtype, expected_dtype, sign):
    """
    Test that the wrapping of the Longitude value range in radians works
    in both float32 and float64.
    """
    # This prevents upcasting to float64 as sign * value would do.
    if sign < 0:
        value = -value
        expected_value = -expected_value

    result = Longitude(value, u.rad, dtype=dtype)
    assert result.value == expected_value
    assert result.dtype == expected_dtype
    assert result.unit == u.rad


@pytest.mark.parametrize("sign", (-1, 1))
@pytest.mark.parametrize(
    "value,expected_value,dtype,expected_dtype",
    [
        (np.pi / 2, np.pi / 2, None, np.float64),
        (np.pi / 2, np.pi / 2, np.float64, np.float64),
        (np.float32(np.pi / 2), np.float32(np.pi / 2), None, np.float32),
        (np.float32(np.pi / 2), np.float32(np.pi / 2), np.float32, np.float32),
        # these cases would require coercing the float32 value to the float64 value
        # making validate have side effects, so it's not implemented for now
        # (np.float32(np.pi / 2), np.pi / 2, np.float64, np.float64),
        # (np.float32(-np.pi / 2), -np.pi / 2, np.float64, np.float64),
    ],
)
def test_latitude_limits(value, expected_value, dtype, expected_dtype, sign):
    """
    Test that the validation of the Latitude value range in radians works
    in both float32 and float64.

    As discussed in issue #13708, before, the float32 representation of pi/2
    was rejected as invalid because the comparison always used the float64
    representation.
    """
    # this prevents upcasting to float64 as sign * value would do
    if sign < 0:
        value = -value
        expected_value = -expected_value

    result = Latitude(value, u.rad, dtype=dtype)
    assert result.value == expected_value
    assert result.dtype == expected_dtype
    assert result.unit == u.rad


@pytest.mark.parametrize(
    "value,dtype",
    [
        (0.50001 * np.pi, np.float32),
        (np.float32(0.50001 * np.pi), np.float32),
        (np.float32(-0.50001 * np.pi), np.float32),
        (0.50001 * np.pi, np.float64),
    ],
)
def test_latitude_out_of_limits(value, dtype):
    """
    Test that values slightly larger than pi/2 are rejected for different dtypes.
    Test cases for issue #13708
    """
    with pytest.raises(
        ValueError,
        match=r"Latitude angle\(s\) must be within -90 deg "
        r"<= angle <= 90 deg, got -?90.001\d* deg",
    ):
        Latitude(value, u.rad, dtype=dtype)


def test_angle_pickle_to_string():
    """
    Ensure that after pickling we can still do to_string on hourangle.

    Regression test for gh-13923.
    """
    angle = Angle(0.25 * u.hourangle)
    expected = angle.to_string()
    via_pickle = pickle.loads(pickle.dumps(angle))
    via_pickle_string = via_pickle.to_string()  # This used to fail.
    assert via_pickle_string == expected



"""
Tests the Angle string formatting capabilities.  SkyCoord formatting is in
test_sky_coord
"""

import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import Angle


def test_to_string_precision():
    # There are already some tests in test_api.py, but this is a regression
    # test for the bug in issue #1319 which caused incorrect formatting of the
    # seconds for precision=0

    angle = Angle(-1.23456789, unit=u.degree)

    assert angle.to_string(precision=3) == "-1d14m04.444s"
    assert angle.to_string(precision=1) == "-1d14m04.4s"
    assert angle.to_string(precision=0) == "-1d14m04s"

    angle2 = Angle(-1.23456789, unit=u.hourangle)

    assert angle2.to_string(precision=3, unit=u.hour) == "-1h14m04.444s"
    assert angle2.to_string(precision=1, unit=u.hour) == "-1h14m04.4s"
    assert angle2.to_string(precision=0, unit=u.hour) == "-1h14m04s"

    # Regression test for #7141
    angle3 = Angle(-0.5, unit=u.degree)
    assert angle3.to_string(precision=0, fields=3) == "-0d30m00s"
    assert angle3.to_string(precision=0, fields=2) == "-0d30m"
    assert angle3.to_string(precision=0, fields=1) == "-1d"


def test_to_string_decimal():
    # There are already some tests in test_api.py, but this is a regression
    # test for the bug in issue #1323 which caused decimal formatting to not
    # work

    angle1 = Angle(2.0, unit=u.degree)

    assert angle1.to_string(decimal=True, precision=3) == "2.000"
    assert angle1.to_string(decimal=True, precision=1) == "2.0"
    assert angle1.to_string(decimal=True, precision=0) == "2"

    angle2 = Angle(3.0, unit=u.hourangle)

    assert angle2.to_string(decimal=True, precision=3) == "3.000"
    assert angle2.to_string(decimal=True, precision=1) == "3.0"
    assert angle2.to_string(decimal=True, precision=0) == "3"

    angle3 = Angle(4.0, unit=u.radian)

    assert angle3.to_string(decimal=True, precision=3) == "4.000"
    assert angle3.to_string(decimal=True, precision=1) == "4.0"
    assert angle3.to_string(decimal=True, precision=0) == "4"


@pytest.mark.parametrize("sep", [":", ":.", "dms", "hms"])
@pytest.mark.parametrize(
    "angle",
    [
        Angle(np.pi / 12, "rad"),
        Angle(15, "deg"),
        Angle(15, "hourangle"),
    ],
)
def test_angle_to_string_decimal_with_sep_error(angle, sep):
    # see https://github.com/astropy/astropy/pull/16085#discussion_r1501177163
    with pytest.raises(
        ValueError,
        match=rf"With decimal=True, separator cannot be used \(got sep='{sep}'\)",
    ):
        angle.to_string(sep=sep, decimal=True)


def test_to_string_formats():
    a = Angle(1.113355, unit=u.deg)
    latex_str = r"$1^\circ06{}^\prime48.078{}^{\prime\prime}$"
    assert a.to_string(format="latex") == latex_str
    assert a.to_string(format="latex_inline") == latex_str
    assert a.to_string(format="unicode") == "1°06′48.078″"

    a = Angle(1.113355, unit=u.hour)
    latex_str = r"$1^{\mathrm{h}}06^{\mathrm{m}}48.078^{\mathrm{s}}$"
    assert a.to_string(format="latex") == latex_str
    assert a.to_string(format="latex_inline") == latex_str
    assert a.to_string(format="unicode") == "1ʰ06ᵐ48.078ˢ"

    a = Angle(1.113355, unit=u.radian)
    assert a.to_string(format="latex") == r"$1.11336\;\mathrm{rad}$"
    assert a.to_string(format="latex_inline") == r"$1.11336\;\mathrm{rad}$"
    assert a.to_string(format="unicode") == "1.11336 rad"


def test_to_string_decimal_formats():
    angle1 = Angle(2.0, unit=u.degree)

    assert angle1.to_string(decimal=True, format="generic") == "2 deg"
    assert angle1.to_string(decimal=True, format="latex") == "$2\\mathrm{{}^{\\circ}}$"
    assert angle1.to_string(decimal=True, format="unicode") == "2°"

    angle2 = Angle(3.0, unit=u.hourangle)
    assert angle2.to_string(decimal=True, format="generic") == "3 hourangle"
    assert angle2.to_string(decimal=True, format="latex") == "$3\\mathrm{{}^{h}}$"
    assert angle2.to_string(decimal=True, format="unicode") == "3ʰ"

    angle3 = Angle(4.0, unit=u.radian)

    assert angle3.to_string(decimal=True, format="generic") == "4 rad"
    assert angle3.to_string(decimal=True, format="latex") == "$4\\;\\mathrm{rad}$"
    assert angle3.to_string(decimal=True, format="unicode") == "4 rad"

    with pytest.raises(ValueError, match="Unknown format"):
        angle3.to_string(decimal=True, format="myformat")


def test_to_string_fields():
    a = Angle(1.113355, unit=u.deg)
    assert a.to_string(fields=1) == r"1d"
    assert a.to_string(fields=2) == r"1d07m"
    assert a.to_string(fields=3) == r"1d06m48.078s"


def test_to_string_padding():
    a = Angle(0.5653, unit=u.deg)
    assert a.to_string(unit="deg", sep=":", pad=True) == r"00:33:55.08"

    # Test to make sure negative angles are padded correctly
    a = Angle(-0.5653, unit=u.deg)
    assert a.to_string(unit="deg", sep=":", pad=True) == r"-00:33:55.08"


def test_sexagesimal_rounding_up():
    a = Angle(359.999999999999, unit=u.deg)

    assert a.to_string(precision=None) == "360d00m00s"
    assert a.to_string(precision=4) == "360d00m00.0000s"
    assert a.to_string(precision=5) == "360d00m00.00000s"
    assert a.to_string(precision=6) == "360d00m00.000000s"
    assert a.to_string(precision=7) == "360d00m00.0000000s"
    assert a.to_string(precision=8) == "360d00m00.00000000s"
    assert a.to_string(precision=9) == "359d59m59.999999996s"

    a = Angle(3.999999, unit=u.deg)
    assert a.to_string(fields=2, precision=None) == "4d00m"
    assert a.to_string(fields=2, precision=1) == "4d00m"
    assert a.to_string(fields=2, precision=5) == "4d00m"
    assert a.to_string(fields=1, precision=1) == "4d"
    assert a.to_string(fields=1, precision=5) == "4d"


def test_to_string_scalar():
    a = Angle(1.113355, unit=u.deg)
    assert isinstance(a.to_string(), str)


def test_to_string_radian_with_precision():
    """
    Regression test for a bug that caused ``to_string`` to crash for angles in
    radians when specifying the precision.
    """

    # Check that specifying the precision works
    a = Angle(3.0, unit=u.rad)
    assert a.to_string(precision=3, sep="fromunit") == "3.000 rad"


def test_sexagesimal_round_down():
    a1 = Angle(1, u.deg).to(u.hourangle)
    a2 = Angle(2, u.deg)
    assert a1.to_string() == "0h04m00s"
    assert a2.to_string() == "2d00m00s"


def test_to_string_fields_colon():
    a = Angle(1.113355, unit=u.deg)
    assert a.to_string(fields=2, sep=":") == "1:07"
    assert a.to_string(fields=3, sep=":") == "1:06:48.078"
    assert a.to_string(fields=1, sep=":") == "1"



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy.stats import RegularEvents, bayesian_blocks


def test_single_change_point(rseed=0):
    rng = np.random.default_rng(rseed)
    x = np.concatenate([rng.random(100), 1 + rng.random(200)])

    bins = bayesian_blocks(x)

    assert len(bins) == 3
    assert_allclose(bins[1], 0.927289, rtol=0.02)


def test_duplicate_events(rseed=0):
    rng = np.random.default_rng(rseed)
    t = rng.random(100)
    t[80:] = t[:20]

    # Using int array as a regression test for gh-6877
    x = np.ones(t.shape, dtype=int)
    x[:20] += 1

    bins1 = bayesian_blocks(t)
    bins2 = bayesian_blocks(t[:80], x[:80])

    assert_allclose(bins1, bins2)


def test_measures_fitness_homoscedastic(rseed=0):
    rng = np.random.default_rng(rseed)
    t = np.linspace(0, 1, 11)
    x = np.exp(-0.5 * (t - 0.5) ** 2 / 0.01**2)
    sigma = 0.05
    x = x + sigma * rng.standard_normal(len(x))

    bins = bayesian_blocks(t, x, sigma, fitness="measures")

    assert_allclose(bins, [0, 0.45, 0.55, 1])


def test_measures_fitness_heteroscedastic():
    rng = np.random.default_rng(1)
    t = np.linspace(0, 1, 11)
    x = np.exp(-0.5 * (t - 0.5) ** 2 / 0.01**2)
    sigma = 0.02 + 0.02 * rng.random(len(x))
    x = x + sigma * rng.standard_normal(len(x))

    bins = bayesian_blocks(t, x, sigma, fitness="measures")

    assert_allclose(bins, [0, 0.45, 0.55, 1])


def test_regular_events():
    rng = np.random.default_rng(1234)
    dt = 0.01
    steps = np.concatenate(
        [np.unique(rng.integers(0, 500, 100)), np.unique(rng.integers(500, 1000, 200))]
    )
    t = dt * steps

    # string fitness
    bins1 = bayesian_blocks(t, fitness="regular_events", dt=dt)
    assert len(bins1) == 3
    assert_allclose(bins1[1], 5, rtol=0.05)

    # class name fitness
    bins2 = bayesian_blocks(t, fitness=RegularEvents, dt=dt)
    assert_allclose(bins1, bins2)

    # class instance fitness
    bins3 = bayesian_blocks(t, fitness=RegularEvents(dt=dt))
    assert_allclose(bins1, bins3)


def test_errors():
    rng = np.random.default_rng(0)
    t = rng.random(100)

    # x must be integer or None for events
    with pytest.raises(ValueError):
        bayesian_blocks(t, fitness="events", x=t)

    # x must be binary for regular events
    with pytest.raises(ValueError):
        bayesian_blocks(t, fitness="regular_events", x=10 * t, dt=1)

    # x must be specified for measures
    with pytest.raises(ValueError):
        bayesian_blocks(t, fitness="measures")

    # sigma cannot be specified without x
    with pytest.raises(ValueError):
        bayesian_blocks(t, fitness="events", sigma=0.5)

    # length of x must match length of t
    with pytest.raises(ValueError):
        bayesian_blocks(t, fitness="measures", x=t[:-1])

    # repeated values in t fail when x is specified
    t2 = t.copy()
    t2[1] = t2[0]
    with pytest.raises(ValueError):
        bayesian_blocks(t2, fitness="measures", x=t)

    # sigma must be broadcastable with x
    with pytest.raises(ValueError):
        bayesian_blocks(t, fitness="measures", x=t, sigma=t[:-1])


def test_fitness_function_results():
    """Test results for several fitness functions"""
    rng = np.random.default_rng(42)

    # Event Data
    t = rng.standard_normal(100)
    edges = bayesian_blocks(t, fitness="events")
    assert_allclose(edges, [-1.95103519, -1.01861547, 0.95442154, 2.1416476])

    # Event data with repeats
    t[80:] = t[:20]
    edges = bayesian_blocks(t, fitness="events", p0=0.01)
    assert_allclose(edges, [-1.95103519, -1.08663566, 1.17575682, 2.1416476])

    # Regular event data
    dt = 0.01
    t = dt * np.arange(1000)
    x = np.zeros(len(t))
    N = len(t) // 10
    x[rng.integers(0, len(t), N)] = 1
    x[rng.integers(0, len(t) // 2, N)] = 1
    edges = bayesian_blocks(t, x, fitness="regular_events", dt=dt)
    assert_allclose(edges, [0, 4.365, 4.995, 9.99])

    # Measured point data with errors
    t = 100 * rng.random(20)
    x = np.exp(-0.5 * (t - 50) ** 2)
    sigma = 0.1
    x_obs = x + sigma * rng.standard_normal(len(x))
    edges = bayesian_blocks(t, x_obs, sigma, fitness="measures")
    expected = [1.39362877, 44.30811196, 49.46626158, 54.37232704, 92.7562551]
    assert_allclose(edges, expected)

    # Optional arguments are passed (p0)
    p0_sel = 0.05
    edges = bayesian_blocks(t, x_obs, sigma, fitness="measures", p0=p0_sel)
    assert_allclose(edges, expected)

    # Optional arguments are passed (ncp_prior)
    ncp_prior_sel = 4 - np.log(73.53 * p0_sel * (len(t) ** -0.478))
    edges = bayesian_blocks(
        t, x_obs, sigma, fitness="measures", ncp_prior=ncp_prior_sel
    )
    assert_allclose(edges, expected)

    # Optional arguments are passed (gamma)
    gamma_sel = np.exp(-ncp_prior_sel)
    edges = bayesian_blocks(t, x_obs, sigma, fitness="measures", gamma=gamma_sel)
    assert_allclose(edges, expected)


def test_zero_change_points(rseed=0):
    """
    Ensure that edges contains both endpoints when there are no change points
    """
    np.random.seed(rseed)
    # Using the failed edge case from
    # https://github.com/astropy/astropy/issues/8558
    values = np.array([1, 1, 1, 1, 1, 1, 1, 1, 2])
    bins = bayesian_blocks(values)
    assert values.min() == bins[0]
    assert values.max() == bins[-1]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module tests some of the methods related to YAML serialization.
"""

from io import StringIO

import numpy as np
import pytest
from yaml import SafeDumper

import astropy.units as u
from astropy.coordinates import (
    Angle,
    CartesianDifferential,
    CartesianRepresentation,
    EarthLocation,
    Latitude,
    Longitude,
    SkyCoord,
    SphericalCosLatDifferential,
    SphericalDifferential,
    SphericalRepresentation,
    UnitSphericalRepresentation,
)
from astropy.coordinates.tests.test_representation import representation_equal
from astropy.io.misc.yaml import AstropyDumper, dump, load, load_all
from astropy.table import QTable, SerializedColumn
from astropy.time import Time


@pytest.mark.parametrize(
    "c",
    [
        True,
        np.uint8(8),
        np.int16(4),
        np.int32(1),
        np.int64(3),
        np.int64(2**63 - 1),
        2.0,
        np.float64(),
        3 + 4j,
        np.complex64(3 + 4j),
        np.complex128(1.0 - 2**-52 + 1j * (1.0 - 2**-52)),
    ],
)
@pytest.mark.usefixtures("without_legacy_printoptions")
def test_numpy_types(c):
    cy = load(dump(c))
    assert c == cy


@pytest.mark.parametrize("c", [float("inf"), float("-inf"), np.inf, -np.inf])
def test_astropydumper_represent_float_override(c):
    # AstropyDumper overrides SafeDumper.represent_float, which has multiple
    # branches, not all of which are intended to deviate.
    # Check that the subclass behave as its parent in these cases.
    d1 = SafeDumper(stream=StringIO())
    d2 = AstropyDumper(stream=StringIO())
    assert d2.represent_float(c).value == d1.represent_float(c).value


@pytest.mark.parametrize(
    "c", [u.m, u.m / u.s, u.hPa, u.dimensionless_unscaled, u.Unit("m, (cm, um)")]
)
def test_unit(c):
    cy = load(dump(c))
    if isinstance(c, (u.CompositeUnit, u.StructuredUnit)):
        assert c == cy
    else:
        assert c is cy


@pytest.mark.parametrize("c", [u.Unit("bakers_dozen", 13 * u.one), u.def_unit("magic")])
def test_custom_unit(c):
    s = dump(c)
    with pytest.warns(u.UnitsWarning, match=f"'{c!s}' did not parse") as w:
        cy = load(s)
    assert len(w) == 1
    assert isinstance(cy, u.UnrecognizedUnit)
    assert str(cy) == str(c)

    with u.add_enabled_units(c):
        cy2 = load(s)
        assert cy2 is c


@pytest.mark.parametrize(
    "c",
    [
        Angle("1 2 3", unit="deg"),
        Longitude("1 2 3", unit="deg"),
        Latitude("1 2 3", unit="deg"),
        [[1], [3]] * u.m,
        np.array([[1, 2], [3, 4]], order="F"),
        np.array([[1, 2], [3, 4]], order="C"),
        np.array([1, 2, 3, 4])[::2],
        np.array([(1.0, 2), (3.0, 4)], dtype="f8,i4"),  # array with structured dtype.
        np.array((1.0, 2), dtype="f8,i4"),  # array scalar with structured dtype.
        np.array((1.0, 2), dtype="f8,i4")[()],  # numpy void.
        np.array((1.0, 2.0), dtype="f8,f8") * u.s,  # Quantity structured scalar.
        [
            ((1.0, 2.0, 3.0), (4.0, 5.0, 6.0)),  # Quantity with structured unit.
            ((11.0, 12.0, 13.0), (14.0, 15.0, 16.0)),
        ]
        * u.Unit("m, m/s"),
        np.array(
            [
                ((1.0, 2.0, 3.0), (4.0, 5.0, 6.0)),
                ((11.0, 12.0, 13.0), (14.0, 15.0, 16.0)),
            ],
            dtype=[("p", "3f8"), ("v", "3f8")],
        )
        * u.Unit("m, m/s"),
    ],
)
def test_ndarray_subclasses(c):
    cy = load(dump(c))

    assert np.all(c == cy)
    assert c.shape == cy.shape
    assert c.dtype == cy.dtype
    assert type(c) is type(cy)

    cc = "C_CONTIGUOUS"
    fc = "F_CONTIGUOUS"
    if c.flags[cc] or c.flags[fc]:
        assert c.flags[cc] == cy.flags[cc]
        assert c.flags[fc] == cy.flags[fc]
    else:
        # Original was not contiguous but round-trip version
        # should be c-contig.
        assert cy.flags[cc]

    if hasattr(c, "unit"):
        assert c.unit == cy.unit


def compare_coord(c, cy):
    assert c.shape == cy.shape
    assert c.frame.name == cy.frame.name

    assert list(c.frame_attributes) == list(cy.frame_attributes)
    for attr in c.frame_attributes:
        assert getattr(c, attr) == getattr(cy, attr)

    assert list(c.representation_component_names) == list(
        cy.representation_component_names
    )
    for name in c.representation_component_names:
        assert np.all(getattr(c, attr) == getattr(cy, attr))


@pytest.mark.parametrize("frame", ["fk4", "altaz"])
def test_skycoord(frame):
    c = SkyCoord(
        [[1, 2], [3, 4]],
        [[5, 6], [7, 8]],
        unit="deg",
        frame=frame,
        obstime=Time("2016-01-02"),
        location=EarthLocation(1000, 2000, 3000, unit=u.km),
    )
    cy = load(dump(c))
    compare_coord(c, cy)


@pytest.mark.parametrize(
    "rep",
    [
        CartesianRepresentation(1 * u.m, 2.0 * u.m, 3.0 * u.m),
        SphericalRepresentation(
            [[1, 2], [3, 4]] * u.deg, [[5, 6], [7, 8]] * u.deg, 10 * u.pc
        ),
        UnitSphericalRepresentation(0 * u.deg, 10 * u.deg),
        SphericalCosLatDifferential(
            [[1.0], [2.0]] * u.mas / u.yr,
            [4.0, 5.0] * u.mas / u.yr,
            [[[10]], [[20]]] * u.km / u.s,
        ),
        CartesianDifferential([10, 20, 30] * u.km / u.s),
        CartesianRepresentation(
            [1, 2, 3] * u.m,
            differentials=CartesianDifferential([10, 20, 30] * u.km / u.s),
        ),
        SphericalRepresentation(
            [[1, 2], [3, 4]] * u.deg,
            [[5, 6], [7, 8]] * u.deg,
            10 * u.pc,
            differentials={
                "s": SphericalDifferential(
                    [[0.0, 1.0], [2.0, 3.0]] * u.mas / u.yr,
                    [[4.0, 5.0], [6.0, 7.0]] * u.mas / u.yr,
                    10 * u.km / u.s,
                )
            },
        ),
    ],
)
def test_representations(rep):
    rrep = load(dump(rep))
    assert np.all(representation_equal(rrep, rep))


def _get_time():
    t = Time(
        [[1], [2]], format="cxcsec", location=EarthLocation(1000, 2000, 3000, unit=u.km)
    )
    t.format = "iso"
    t.precision = 5
    t.delta_ut1_utc = np.array([[3.0], [4.0]])
    t.delta_tdb_tt = np.array([[5.0], [6.0]])
    t.out_subfmt = "date_hm"

    return t


def compare_time(t, ty):
    assert type(t) is type(ty)
    assert np.all(t == ty)
    for attr in (
        "shape",
        "jd1",
        "jd2",
        "format",
        "scale",
        "precision",
        "in_subfmt",
        "out_subfmt",
        "location",
        "delta_ut1_utc",
        "delta_tdb_tt",
    ):
        assert np.all(getattr(t, attr) == getattr(ty, attr))


def test_time():
    t = _get_time()
    ty = load(dump(t))
    compare_time(t, ty)


def test_timedelta():
    t = _get_time()
    dt = t - t + 0.1234556 * u.s
    dty = load(dump(dt))

    assert type(dt) is type(dty)
    for attr in ("shape", "jd1", "jd2", "format", "scale"):
        assert np.all(getattr(dt, attr) == getattr(dty, attr))


def test_serialized_column():
    sc = SerializedColumn({"name": "hello", "other": 1, "other2": 2.0})
    scy = load(dump(sc))

    assert sc == scy


def test_load_all():
    t = _get_time()
    unit = u.m / u.s
    c = SkyCoord(
        [[1, 2], [3, 4]],
        [[5, 6], [7, 8]],
        unit="deg",
        frame="fk4",
        obstime=Time("2016-01-02"),
        location=EarthLocation(1000, 2000, 3000, unit=u.km),
    )

    # Make a multi-document stream
    out = "---\n" + dump(t) + "---\n" + dump(unit) + "---\n" + dump(c)

    ty, unity, cy = list(load_all(out))

    compare_time(t, ty)
    compare_coord(c, cy)
    assert unity == unit


def test_ecsv_astropy_objects_in_meta():
    """
    Test that astropy core objects in ``meta`` are serialized.
    """
    t = QTable([[1, 2] * u.m, [4, 5]], names=["a", "b"])
    tm = _get_time()
    c = SkyCoord(
        [[1, 2], [3, 4]],
        [[5, 6], [7, 8]],
        unit="deg",
        frame="fk4",
        obstime=Time("2016-01-02"),
        location=EarthLocation(1000, 2000, 3000, unit=u.km),
    )
    unit = u.m / u.s

    t.meta = {"tm": tm, "c": c, "unit": unit}
    out = StringIO()
    t.write(out, format="ascii.ecsv")
    t2 = QTable.read(out.getvalue(), format="ascii.ecsv")

    compare_time(tm, t2.meta["tm"])
    compare_coord(c, t2.meta["c"])
    assert t2.meta["unit"] == unit


def test_yaml_dump_of_object_arrays_fail():
    """Test that dumping and loading object arrays fails."""
    with pytest.raises(TypeError, match="cannot serialize"):
        dump(np.array([1, 2, 3], dtype=object))


def test_yaml_load_of_object_arrays_fail():
    """Test that dumping and loading object arrays fails.

    The string to load was obtained by suppressing the exception and dumping
    ``np.array([1, 2, 3], dtype=object)`` to a yaml file.
    """
    with pytest.raises(TypeError, match="cannot load numpy array"):
        load(
            """!numpy.ndarray
            buffer: !!binary |
              WndBQUFISUFBQUJwQUFBQQ==
            dtype: object
            order: C
            shape: !!python/tuple [3]"""
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from functools import cache

from astropy import units as u
from astropy.coordinates.attributes import CoordinateAttribute, QuantityAttribute
from astropy.coordinates.baseframe import BaseCoordinateFrame, frame_transform_graph
from astropy.coordinates.matrix_utilities import matrix_transpose, rotation_matrix
from astropy.coordinates.transformations import (
    DynamicMatrixTransform,
    FunctionTransform,
)


@cache
def make_skyoffset_cls(framecls):
    """
    Create a new class that is the sky offset frame for a specific class of
    origin frame. If such a class has already been created for this frame, the
    same class will be returned.

    The new class will always have component names for spherical coordinates of
    ``lon``/``lat``.

    Parameters
    ----------
    framecls : `~astropy.coordinates.BaseCoordinateFrame` subclass
        The class to create the SkyOffsetFrame of.

    Returns
    -------
    skyoffsetframecls : class
        The class for the new skyoffset frame.

    Notes
    -----
    This function is necessary because Astropy's frame transformations depend
    on connection between specific frame *classes*.  So each type of frame
    needs its own distinct skyoffset frame class.  This function generates
    just that class, as well as ensuring that only one example of such a class
    actually gets created in any given python session.
    """
    # Create a new SkyOffsetFrame subclass for this frame class.
    name = "SkyOffset" + framecls.__name__
    _SkyOffsetFramecls = type(
        name,
        (SkyOffsetFrame, framecls),
        {
            "origin": CoordinateAttribute(
                frame=framecls, default=None, doc="The origin of the offset frame"
            ),
            # The following two have to be done because otherwise we use the
            # defaults of SkyOffsetFrame set by BaseCoordinateFrame.
            "_default_representation": framecls._default_representation,
            "_default_differential": framecls._default_differential,
            "__doc__": SkyOffsetFrame.__doc__,
        },
    )

    @frame_transform_graph.transform(
        FunctionTransform, _SkyOffsetFramecls, _SkyOffsetFramecls
    )
    def skyoffset_to_skyoffset(from_skyoffset_coord, to_skyoffset_frame):
        """Transform between two skyoffset frames."""
        # This transform goes through the parent frames on each side.
        # from_frame -> from_frame.origin -> to_frame.origin -> to_frame
        tmp_from = from_skyoffset_coord.transform_to(from_skyoffset_coord.origin)
        tmp_to = tmp_from.transform_to(to_skyoffset_frame.origin)
        return tmp_to.transform_to(to_skyoffset_frame)

    @frame_transform_graph.transform(
        DynamicMatrixTransform, framecls, _SkyOffsetFramecls
    )
    def reference_to_skyoffset(reference_frame, skyoffset_frame):
        """Convert a reference coordinate to an sky offset frame."""
        # Define rotation matrices along the position angle vector, and
        # relative to the origin.
        origin = skyoffset_frame.origin.spherical
        return (
            rotation_matrix(-skyoffset_frame.rotation, "x")
            @ rotation_matrix(-origin.lat, "y")
            @ rotation_matrix(origin.lon, "z")
        )

    @frame_transform_graph.transform(
        DynamicMatrixTransform, _SkyOffsetFramecls, framecls
    )
    def skyoffset_to_reference(skyoffset_coord, reference_frame):
        """Convert an sky offset frame coordinate to the reference frame."""
        # use the forward transform, but just invert it
        R = reference_to_skyoffset(reference_frame, skyoffset_coord)
        # transpose is the inverse because R is a rotation matrix
        return matrix_transpose(R)

    return _SkyOffsetFramecls


class SkyOffsetFrame(BaseCoordinateFrame):
    """
    A frame which is relative to some specific position and oriented to match
    its frame.

    SkyOffsetFrames always have component names for spherical coordinates
    of ``lon``/``lat``, *not* the component names for the frame of ``origin``.

    This is useful for calculating offsets and dithers in the frame of the sky
    relative to an arbitrary position. Coordinates in this frame are both centered on the position specified by the
    ``origin`` coordinate, *and* they are oriented in the same manner as the
    ``origin`` frame.  E.g., if ``origin`` is `~astropy.coordinates.ICRS`, this
    object's ``lat`` will be pointed in the direction of Dec, while ``lon``
    will point in the direction of RA.

    For more on skyoffset frames, see :ref:`astropy:astropy-skyoffset-frames`.

    Parameters
    ----------
    representation : `~astropy.coordinates.BaseRepresentation` or None
        A representation object or None to have no data (or use the other keywords)
    origin : coordinate-like
        The coordinate which specifies the origin of this frame. Note that this
        origin is used purely for on-sky location/rotation.  It can have a
        ``distance`` but it will not be used by this ``SkyOffsetFrame``.
    rotation : angle-like
        The final rotation of the frame about the ``origin``. The sign of
        the rotation is the left-hand rule.  That is, an object at a
        particular position angle in the un-rotated system will be sent to
        the positive latitude (z) direction in the final frame.


    Notes
    -----
    ``SkyOffsetFrame`` is a factory class.  That is, the objects that it
    yields are *not* actually objects of class ``SkyOffsetFrame``.  Instead,
    distinct classes are created on-the-fly for whatever the frame class is
    of ``origin``.
    """

    rotation = QuantityAttribute(
        default=0, unit=u.deg, doc="The rotation angle for the frame orientation"
    )
    origin = CoordinateAttribute(
        default=None, frame=None, doc="The origin of the offset frame"
    )

    def __new__(cls, *args, **kwargs):
        # We don't want to call this method if we've already set up
        # an skyoffset frame for this class.
        if not (issubclass(cls, SkyOffsetFrame) and cls is not SkyOffsetFrame):
            # We get the origin argument, and handle it here.
            try:
                origin_frame = kwargs["origin"]
            except KeyError:
                raise TypeError(
                    "Can't initialize a SkyOffsetFrame without origin= keyword."
                )
            if hasattr(origin_frame, "frame"):
                origin_frame = origin_frame.frame
            newcls = make_skyoffset_cls(origin_frame.__class__)
            return newcls.__new__(newcls, *args, **kwargs)

        # http://stackoverflow.com/questions/19277399/why-does-object-new-work-differently-in-these-three-cases
        # See above for why this is necessary. Basically, because some child
        # may override __new__, we must override it here to never pass
        # arguments to the object.__new__ method.
        if super().__new__ is object.__new__:
            return super().__new__(cls)
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.origin is not None and not self.origin.has_data:
            raise ValueError("The origin supplied to SkyOffsetFrame has no data.")
        if self.has_data:
            self._set_skyoffset_data_lon_wrap_angle(self.data)

    @staticmethod
    def _set_skyoffset_data_lon_wrap_angle(data):
        if hasattr(data, "lon"):
            data.lon.wrap_angle = 180.0 * u.deg
        return data

    def represent_as(self, base, s="base", in_frame_units=False):
        """
        Ensure the wrap angle for any spherical
        representations.
        """
        data = super().represent_as(base, s, in_frame_units=in_frame_units)
        self._set_skyoffset_data_lon_wrap_angle(data)
        return data

    def __reduce__(self):
        return (_skyoffset_reducer, (self.origin,), self.__dict__)


def _skyoffset_reducer(origin):
    return SkyOffsetFrame.__new__(SkyOffsetFrame, origin=origin)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Script support for validating the WCS keywords in a FITS file.
"""


def main(args=None):
    import argparse

    from . import wcs

    parser = argparse.ArgumentParser(
        description=(
            "Check the WCS keywords in a FITS file for compliance against the standards"
        )
    )
    parser.add_argument("filename", nargs=1, help="Path to FITS file to check")
    args = parser.parse_args(args)

    print(wcs.validate(args.filename[0]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Helpers for letting numpy functions interact with Masked arrays.

The module supplies helper routines for numpy functions that propagate
masks appropriately, for use in the ``__array_function__``
implementation of `~astropy.utils.masked.MaskedNDArray`.  They are not
very useful on their own, but the ones with docstrings are included in
the documentation so that there is a place to find out how the mask is
interpreted.

"""

import warnings

import numpy as np

from astropy.units.quantity_helper.function_helpers import FunctionAssigner
from astropy.utils.compat import NUMPY_LT_1_24, NUMPY_LT_2_0, NUMPY_LT_2_1

if NUMPY_LT_2_0:
    import numpy.core as np_core
    from numpy.lib.function_base import (
        _check_interpolation_as_method,
        _quantile_is_valid,
        _ureduce,
    )
else:
    import numpy._core as np_core
    from numpy.lib._function_base_impl import (
        _check_interpolation_as_method,
        _quantile_is_valid,
        _ureduce,
    )


# This module should not really be imported, but we define __all__
# such that sphinx can typeset the functions with docstrings.
# The latter are added to __all__ at the end.
__all__ = [
    "MASKED_SAFE_FUNCTIONS",
    "APPLY_TO_BOTH_FUNCTIONS",
    "DISPATCHED_FUNCTIONS",
    "UNSUPPORTED_FUNCTIONS",
]


MASKED_SAFE_FUNCTIONS = set()
"""Set of functions that work fine on Masked classes already.

Most of these internally use `numpy.ufunc` or other functions that
are already covered.
"""

APPLY_TO_BOTH_FUNCTIONS = {}
"""Dict of functions that should apply to both data and mask.

The `dict` is keyed by the numpy function and the values are functions
that take the input arguments of the numpy function and organize these
for passing the data and mask to the numpy function.

Returns
-------
data_args : tuple
    Arguments to pass on to the numpy function for the unmasked data.
mask_args : tuple
    Arguments to pass on to the numpy function for the masked data.
kwargs : dict
    Keyword arguments to pass on for both unmasked data and mask.
out : `~astropy.utils.masked.Masked` instance or None
    Optional instance in which to store the output.

Raises
------
NotImplementedError
   When an arguments is masked when it should not be or vice versa.
"""

DISPATCHED_FUNCTIONS = {}
"""Dict of functions that provide the numpy function's functionality.

These are for more complicated versions where the numpy function itself
cannot easily be used.  It should return either the result of the
function, or a tuple consisting of the unmasked result, the mask for the
result and a possible output instance.

It should raise `NotImplementedError` if one of the arguments is masked
when it should not be or vice versa.
"""

UNSUPPORTED_FUNCTIONS = set()
"""Set of numpy functions that are not supported for masked arrays.

For most, masked input simply makes no sense, but for others it may have
been lack of time.  Issues or PRs for support for functions are welcome.
"""

# Almost all from np.core.fromnumeric defer to methods so are OK.
MASKED_SAFE_FUNCTIONS |= {
    getattr(np, name)
    for name in np_core.fromnumeric.__all__
    if name not in {"choose", "put", "resize", "searchsorted", "where", "alen", "ptp"}
}
MASKED_SAFE_FUNCTIONS |= {
    # built-in from multiarray
    np.may_share_memory, np.can_cast, np.min_scalar_type, np.result_type,
    np.shares_memory,
    # np.core.arrayprint
    np.array_repr,
    # np.core.function_base
    np.linspace, np.logspace, np.geomspace,
    # np.core.numeric
    np.isclose, np.allclose, np.flatnonzero, np.argwhere,
    # np.core.shape_base
    np.atleast_1d, np.atleast_2d, np.atleast_3d, np.stack, np.hstack, np.vstack,
    # np.lib._function_base_impl
    np.average, np.diff, np.extract, np.meshgrid, np.gradient,
    # np.lib.index_tricks
    np.diag_indices_from, np.triu_indices_from, np.tril_indices_from,
    np.fill_diagonal,
    # np.lib.shape_base
    np.column_stack, np.dstack,
    np.array_split, np.split, np.hsplit, np.vsplit, np.dsplit,
    np.expand_dims, np.apply_along_axis, np.kron, np.tile,
    np.take_along_axis, np.put_along_axis,
    # np.lib.type_check (all but nan_to_num)
    np.iscomplexobj, np.isrealobj, np.imag, np.isreal, np.real,
    np.real_if_close, np.common_type,
    # np.lib.ufunclike
    np.fix, np.isneginf, np.isposinf,
    # np.lib._function_base_impl
    np.angle, np.i0,
    # np.lib._arraysetops_impl
    np.intersect1d, np.setxor1d, np.union1d, np.unique,
}  # fmt: skip


if NUMPY_LT_2_0:
    # Safe in < 2.0, because it deferred to the method. Overridden in >= 2.0.
    MASKED_SAFE_FUNCTIONS |= {np.ptp}
    # Removed in numpy 2.0.  Just an alias to vstack.
    MASKED_SAFE_FUNCTIONS |= {np.row_stack}
    # renamed in numpy 2.0
    MASKED_SAFE_FUNCTIONS |= {np.trapz}
if not NUMPY_LT_2_0:
    # new in numpy 2.0
    MASKED_SAFE_FUNCTIONS |= {
        np.astype, np.trapezoid,
        np.unique_all, np.unique_counts, np.unique_inverse, np.unique_values,
    }  # fmt: skip
if not NUMPY_LT_2_1:
    MASKED_SAFE_FUNCTIONS |= {
        np.unstack,
    }  # fmt: skip
IGNORED_FUNCTIONS = {
    # I/O - useless for Masked, since no way to store the mask.
    np.save, np.savez, np.savetxt, np.savez_compressed,
    # Polynomials
    np.poly, np.polyadd, np.polyder, np.polydiv, np.polyfit, np.polyint,
    np.polymul, np.polysub, np.polyval, np.roots, np.vander,
}  # fmt: skip
IGNORED_FUNCTIONS |= {
    np.pad, np.searchsorted, np.digitize,
    np.is_busday, np.busday_count, np.busday_offset,
    # numpy.lib._function_base_impl
    np.cov, np.corrcoef, np.trim_zeros,
    # numpy.core.numeric
    np.correlate, np.convolve,
    # numpy.lib.histograms
    np.histogram, np.histogram2d, np.histogramdd, np.histogram_bin_edges,
    # TODO!!
    np.dot, np.vdot, np.inner, np.tensordot, np.cross,
    np.einsum, np.einsum_path,
}  # fmt: skip

# Explicitly unsupported functions
UNSUPPORTED_FUNCTIONS |= {
    np.unravel_index,
    np.ravel_multi_index,
    np.ix_,
}

# No support for the functions also not supported by Quantity
# (io, polynomial, etc.).
UNSUPPORTED_FUNCTIONS |= IGNORED_FUNCTIONS


apply_to_both = FunctionAssigner(APPLY_TO_BOTH_FUNCTIONS)
dispatched_function = FunctionAssigner(DISPATCHED_FUNCTIONS)


def _get_data_and_masks(*args):
    """Separate out arguments into tuples of data and masks.

    An all-False mask is created if an argument does not have a mask.
    """
    from .core import Masked

    data, masks = Masked._get_data_and_masks(*args)
    masks = tuple(
        m if m is not None else np.zeros(np.shape(d), bool) for d, m in zip(data, masks)
    )
    return data, masks


# Following are simple ufunc-like functions which should just copy the mask.
@dispatched_function
def datetime_as_string(arr, *args, **kwargs):
    return (np.datetime_as_string(arr.unmasked, *args, **kwargs), arr.mask.copy(), None)


@dispatched_function
def sinc(x):
    return np.sinc(x.unmasked), x.mask.copy(), None


@dispatched_function
def iscomplex(x):
    return np.iscomplex(x.unmasked), x.mask.copy(), None


@dispatched_function
def unwrap(p, *args, **kwargs):
    return np.unwrap(p.unmasked, *args, **kwargs), p.mask.copy(), None


@dispatched_function
def nan_to_num(x, copy=True, nan=0.0, posinf=None, neginf=None):
    data = np.nan_to_num(x.unmasked, copy=copy, nan=nan, posinf=posinf, neginf=neginf)
    if copy:
        return (data, x.mask.copy(), None)
    else:
        return (x, None, None)


# Following are simple functions related to shapes, where the same function
# should be applied to the data and the mask.  They cannot all share the
# same helper, because the first arguments have different names.
@apply_to_both(
    helps=(
        {np.copy, np.resize, np.moveaxis, np.rollaxis, np.roll}
        | ({np.asfarray} if NUMPY_LT_2_0 else set())  # noqa: NPY201
    )
)
def masked_a_helper(a, *args, **kwargs):
    data, mask = _get_data_and_masks(a)
    return data + args, mask + args, kwargs, None


@apply_to_both(helps={np.flip, np.flipud, np.fliplr, np.rot90, np.triu, np.tril})
def masked_m_helper(m, *args, **kwargs):
    data, mask = _get_data_and_masks(m)
    return data + args, mask + args, kwargs, None


@apply_to_both(helps={np.diag, np.diagflat})
def masked_v_helper(v, *args, **kwargs):
    data, mask = _get_data_and_masks(v)
    return data + args, mask + args, kwargs, None


@apply_to_both(helps={np.delete})
def masked_arr_helper(array, *args, **kwargs):
    data, mask = _get_data_and_masks(array)
    return data + args, mask + args, kwargs, None


@apply_to_both
def broadcast_to(array, shape, subok=False):
    """Broadcast array to the given shape.

    Like `numpy.broadcast_to`, and applied to both unmasked data and mask.
    Note that ``subok`` is taken to mean whether or not subclasses of
    the unmasked data and mask are allowed, i.e., for ``subok=False``,
    a `~astropy.utils.masked.MaskedNDArray` will be returned.
    """
    data, mask = _get_data_and_masks(array)
    return data, mask, dict(shape=shape, subok=subok), None


@dispatched_function
def outer(a, b, out=None):
    return np.multiply.outer(np.ravel(a), np.ravel(b), out=out), None, None


if not NUMPY_LT_2_0:

    @dispatched_function
    def empty_like(
        prototype, dtype=None, order="K", subok=True, shape=None, *, device=None
    ):
        """Return a new array with the same shape and type as a given array.

        Like `numpy.empty_like`, but will add an empty mask.
        """
        unmasked = np.empty_like(
            prototype.unmasked,
            dtype=dtype,
            order=order,
            subok=subok,
            shape=shape,
            device=device,
        )
        if dtype is not None:
            dtype = (
                np.ma.make_mask_descr(unmasked.dtype)
                if unmasked.dtype.names
                else np.dtype("?")
            )
        mask = np.empty_like(
            prototype.mask,
            dtype=dtype,
            order=order,
            subok=subok,
            shape=shape,
            device=device,
        )

        return unmasked, mask, None

    @dispatched_function
    def zeros_like(a, dtype=None, order="K", subok=True, shape=None, *, device=None):
        """Return an array of zeros with the same shape and type as a given array.

        Like `numpy.zeros_like`, but will add an all-false mask.
        """
        unmasked = np.zeros_like(
            a.unmasked,
            dtype=dtype,
            order=order,
            subok=subok,
            shape=shape,
            device=device,
        )
        return unmasked, False, None

    @dispatched_function
    def ones_like(a, dtype=None, order="K", subok=True, shape=None, *, device=None):
        """Return an array of ones with the same shape and type as a given array.

        Like `numpy.ones_like`, but will add an all-false mask.
        """
        unmasked = np.ones_like(
            a.unmasked,
            dtype=dtype,
            order=order,
            subok=subok,
            shape=shape,
            device=device,
        )
        return unmasked, False, None

    @dispatched_function
    def full_like(
        a, fill_value, dtype=None, order="K", subok=True, shape=None, *, device=None
    ):
        """Return a full array with the same shape and type as a given array.

        Like `numpy.full_like`, but with a mask that is also set.
        If ``fill_value`` is `numpy.ma.masked`, the data will be left unset
        (i.e., as created by `numpy.empty_like`).
        """
        result = np.empty_like(
            a, dtype=dtype, order=order, subok=subok, shape=shape, device=device
        )
        result[...] = fill_value
        return result, None, None

else:

    @dispatched_function
    def empty_like(prototype, dtype=None, order="K", subok=True, shape=None):
        """Return a new array with the same shape and type as a given array.

        Like `numpy.empty_like`, but will add an empty mask.
        """
        unmasked = np.empty_like(
            prototype.unmasked, dtype=dtype, order=order, subok=subok, shape=shape
        )
        if dtype is not None:
            dtype = (
                np.ma.make_mask_descr(unmasked.dtype)
                if unmasked.dtype.names
                else np.dtype("?")
            )
        mask = np.empty_like(
            prototype.mask, dtype=dtype, order=order, subok=subok, shape=shape
        )

        return unmasked, mask, None

    @dispatched_function
    def zeros_like(a, dtype=None, order="K", subok=True, shape=None):
        """Return an array of zeros with the same shape and type as a given array.

        Like `numpy.zeros_like`, but will add an all-false mask.
        """
        unmasked = np.zeros_like(
            a.unmasked, dtype=dtype, order=order, subok=subok, shape=shape
        )
        return unmasked, False, None

    @dispatched_function
    def ones_like(a, dtype=None, order="K", subok=True, shape=None):
        """Return an array of ones with the same shape and type as a given array.

        Like `numpy.ones_like`, but will add an all-false mask.
        """
        unmasked = np.ones_like(
            a.unmasked, dtype=dtype, order=order, subok=subok, shape=shape
        )
        return unmasked, False, None

    @dispatched_function
    def full_like(a, fill_value, dtype=None, order="K", subok=True, shape=None):
        """Return a full array with the same shape and type as a given array.

        Like `numpy.full_like`, but with a mask that is also set.
        If ``fill_value`` is `numpy.ma.masked`, the data will be left unset
        (i.e., as created by `numpy.empty_like`).
        """
        result = np.empty_like(a, dtype=dtype, order=order, subok=subok, shape=shape)
        result[...] = fill_value
        return result, None, None


@dispatched_function
def put(a, ind, v, mode="raise"):
    """Replaces specified elements of an array with given values.

    Like `numpy.put`, but for masked array ``a`` and possibly masked
    value ``v``.  Masked indices ``ind`` are not supported.
    """
    from astropy.utils.masked import Masked

    if isinstance(ind, Masked) or not isinstance(a, Masked):
        raise NotImplementedError

    v_data, v_mask = a._get_data_and_mask(v)
    if v_data is not None:
        np.put(a.unmasked, ind, v_data, mode=mode)
    # v_mask of None will be correctly interpreted as False.
    np.put(a.mask, ind, v_mask, mode=mode)


@dispatched_function
def putmask(a, mask, values):
    """Changes elements of an array based on conditional and input values.

    Like `numpy.putmask`, but for masked array ``a`` and possibly masked
    ``values``.  Masked ``mask`` is not supported.
    """
    from astropy.utils.masked import Masked

    if isinstance(mask, Masked) or not isinstance(a, Masked):
        raise NotImplementedError

    values_data, values_mask = a._get_data_and_mask(values)
    if values_data is not None:
        np.putmask(a.unmasked, mask, values_data)
    np.putmask(a.mask, mask, values_mask)


@dispatched_function
def place(arr, mask, vals):
    """Change elements of an array based on conditional and input values.

    Like `numpy.place`, but for masked array ``a`` and possibly masked
    ``values``.  Masked ``mask`` is not supported.
    """
    from astropy.utils.masked import Masked

    if isinstance(mask, Masked) or not isinstance(arr, Masked):
        raise NotImplementedError

    vals_data, vals_mask = arr._get_data_and_mask(vals)
    if vals_data is not None:
        np.place(arr.unmasked, mask, vals_data)
    np.place(arr.mask, mask, vals_mask)


@dispatched_function
def copyto(dst, src, casting="same_kind", where=True):
    """Copies values from one array to another, broadcasting as necessary.

    Like `numpy.copyto`, but for masked destination ``dst`` and possibly
    masked source ``src``.
    """
    from astropy.utils.masked import Masked

    if not isinstance(dst, Masked) or isinstance(where, Masked):
        raise NotImplementedError

    src_data, src_mask = dst._get_data_and_mask(src)

    if src_data is not None:
        np.copyto(dst.unmasked, src_data, casting=casting, where=where)
    if src_mask is not None:
        np.copyto(dst.mask, src_mask, where=where)


@dispatched_function
def packbits(a, *args, **kwargs):
    result = np.packbits(a.unmasked, *args, **kwargs)
    mask = np.packbits(a.mask, *args, **kwargs).astype(bool)
    return result, mask, None


@dispatched_function
def unpackbits(a, *args, **kwargs):
    result = np.unpackbits(a.unmasked, *args, **kwargs)
    mask = np.zeros(a.shape, dtype="u1")
    mask[a.mask] = 255
    mask = np.unpackbits(mask, *args, **kwargs).astype(bool)
    return result, mask, None


@dispatched_function
def bincount(x, weights=None, minlength=0):
    """Count number of occurrences of each value in array of non-negative ints.

    Like `numpy.bincount`, but masked entries in ``x`` will be skipped.
    Any masked entries in ``weights`` will lead the corresponding bin to
    be masked.
    """
    from astropy.utils.masked import Masked

    if weights is not None:
        weights = np.asanyarray(weights)
    if isinstance(x, Masked) and x.ndim <= 1:
        # let other dimensions lead to errors.
        if weights is not None and weights.ndim == x.ndim:
            weights = weights[~x.mask]
        x = x.unmasked[~x.mask]
    mask = None
    if weights is not None:
        weights, w_mask = Masked._get_data_and_mask(weights)
        if w_mask is not None:
            mask = np.bincount(x, w_mask.astype(int), minlength=minlength).astype(bool)
    result = np.bincount(x, weights, minlength=0)
    return result, mask, None


if NUMPY_LT_2_0:

    @dispatched_function
    def msort(a):
        result = a.copy()
        result.sort(axis=0)
        return result, None, None

else:
    # Used to work via ptp method, but now need to override, otherwise
    # plain reduction is used, which gives different mask.
    @dispatched_function
    def ptp(a, axis=None, out=None, keepdims=np._NoValue):
        if keepdims is np._NoValue:
            keepdims = False
        result = a.max(axis=axis, out=out, keepdims=keepdims)
        result -= a.min(axis=axis, keepdims=keepdims)
        return result, None, None


@dispatched_function
def sort_complex(a):
    # Just a copy of np.lib._function_base_impl.sort_complex, to avoid the asarray.
    b = a.copy()
    b.sort()
    if not issubclass(b.dtype.type, np.complexfloating):  # pragma: no cover
        if b.dtype.char in "bhBH":
            result = b.astype("F")
        elif b.dtype.char == "g":
            result = b.astype("G")
        else:
            result = b.astype("D")
    else:
        result = b

    return result, None, None


@dispatched_function
def concatenate(arrays, axis=0, out=None, dtype=None, casting="same_kind"):
    data, masks = _get_data_and_masks(*arrays)
    if out is None:
        return (
            np.concatenate(data, axis=axis, dtype=dtype, casting=casting),
            np.concatenate(masks, axis=axis),
            None,
        )
    else:
        from astropy.utils.masked import Masked

        if not isinstance(out, Masked):
            raise NotImplementedError
        np.concatenate(masks, out=out.mask, axis=axis)
        np.concatenate(data, out=out.unmasked, axis=axis, dtype=dtype, casting=casting)
        return out, None, None


@apply_to_both
def append(arr, values, axis=None):
    data, masks = _get_data_and_masks(arr, values)
    return data, masks, dict(axis=axis), None


@dispatched_function
def block(arrays):
    # We need to override block since the numpy implementation can take two
    # different paths, one for concatenation, one for creating a large empty
    # result array in which parts are set.  Each assumes array input and
    # cannot be used directly.  Since it would be very costly to inspect all
    # arrays and then turn them back into a nested list, we just copy here the
    # second implementation, np.core.shape_base._block_slicing, since it is
    # shortest and easiest.
    from astropy.utils.masked import Masked

    arrays, list_ndim, result_ndim, final_size = np_core.shape_base._block_setup(arrays)
    shape, slices, arrays = np_core.shape_base._block_info_recursion(
        arrays, list_ndim, result_ndim
    )
    dtype = np.result_type(*[arr.dtype for arr in arrays])
    F_order = all(arr.flags["F_CONTIGUOUS"] for arr in arrays)
    C_order = all(arr.flags["C_CONTIGUOUS"] for arr in arrays)
    order = "F" if F_order and not C_order else "C"
    result = Masked(np.empty(shape=shape, dtype=dtype, order=order))
    for the_slice, arr in zip(slices, arrays):
        result[(Ellipsis,) + the_slice] = arr
    return result, None, None


@dispatched_function
def broadcast_arrays(*args, subok=False):
    """Broadcast arrays to a common shape.

    Like `numpy.broadcast_arrays`, applied to both unmasked data and masks.
    Note that ``subok`` is taken to mean whether or not subclasses of
    the unmasked data and masks are allowed, i.e., for ``subok=False``,
    `~astropy.utils.masked.MaskedNDArray` instances will be returned.
    """
    from .core import Masked

    are_masked = [isinstance(arg, Masked) for arg in args]
    data = [
        (arg.unmasked if is_masked else arg) for arg, is_masked in zip(args, are_masked)
    ]
    results = np.broadcast_arrays(*data, subok=subok)

    return_type = list if NUMPY_LT_2_0 else tuple
    shape = results[0].shape
    masks = [
        (np.broadcast_to(arg.mask, shape, subok=subok) if is_masked else None)
        for arg, is_masked in zip(args, are_masked)
    ]
    results = return_type(
        (Masked(result, mask) if mask is not None else result)
        for (result, mask) in zip(results, masks)
    )
    return results, None, None


@apply_to_both
def insert(arr, obj, values, axis=None):
    """Insert values along the given axis before the given indices.

    Like `numpy.insert` but for possibly masked ``arr`` and ``values``.
    Masked ``obj`` is not supported.
    """
    from astropy.utils.masked import Masked

    if isinstance(obj, Masked) or not isinstance(arr, Masked):
        raise NotImplementedError

    (arr_data, val_data), (arr_mask, val_mask) = _get_data_and_masks(arr, values)
    return ((arr_data, obj, val_data, axis), (arr_mask, obj, val_mask, axis), {}, None)


@dispatched_function
def count_nonzero(a, axis=None, *, keepdims=False):
    """Counts the number of non-zero values in the array ``a``.

    Like `numpy.count_nonzero`, with masked values counted as 0 or `False`.
    """
    filled = a.filled(np.zeros((), a.dtype))
    return np.count_nonzero(filled, axis, keepdims=keepdims), None, None


def _masked_median_1d(a, overwrite_input, keepdims):
    # TODO: need an in-place mask-sorting option.
    unmasked = a.unmasked[~a.mask]
    if unmasked.size:
        return a.from_unmasked(
            np.median(unmasked, overwrite_input=overwrite_input, keepdims=keepdims)
        )
    else:
        return a.from_unmasked(np.zeros_like(a.unmasked, shape=(1,))[0], mask=True)


def _masked_median(a, axis=None, out=None, overwrite_input=False, keepdims=False):
    # As for np.nanmedian, but without a fast option as yet.
    if axis is None or a.ndim == 1:
        part = a.ravel()
        result = _masked_median_1d(part, overwrite_input, keepdims)
    else:
        result = np.apply_along_axis(
            _masked_median_1d, axis, a, overwrite_input, keepdims
        )
    if out is not None:
        out[...] = result
    return result


@dispatched_function
def median(a, axis=None, out=None, overwrite_input=False, keepdims=False):
    from astropy.utils.masked import Masked

    if out is not None and not isinstance(out, Masked):
        raise NotImplementedError

    a = Masked(a)

    if NUMPY_LT_1_24:
        r, k = _ureduce(
            a,
            func=_masked_median,
            axis=axis,
            out=out,
            overwrite_input=overwrite_input,
        )
        result = (r.reshape(k) if keepdims else r) if out is None else out

    else:
        result = _ureduce(
            a,
            func=_masked_median,
            axis=axis,
            out=out,
            overwrite_input=overwrite_input,
            keepdims=keepdims,
        )
    return result, None, None


def _masked_quantile_1d(a, q, **kwargs):
    """
    Private function for rank 1 arrays. Compute quantile ignoring NaNs.
    See nanpercentile for parameter usage.
    """
    unmasked = a.unmasked[~a.mask]
    if unmasked.size:
        if NUMPY_LT_2_0:
            if "weights" in kwargs:
                kwargs.pop("weights")
            result = np.lib.function_base._quantile_unchecked(unmasked, q, **kwargs)
        else:
            result = np.lib._function_base_impl._quantile_unchecked(
                unmasked, q, **kwargs
            )
        return a.from_unmasked(result)
    else:
        return a.from_unmasked(np.zeros_like(a.unmasked, shape=q.shape), True)


def _masked_quantile(a, q, axis=None, out=None, **kwargs):
    # As for np.nanmedian, but without a fast option as yet.
    if axis is None or a.ndim == 1:
        part = a.ravel()
        result = _masked_quantile_1d(part, q, **kwargs)
    else:
        result = np.apply_along_axis(_masked_quantile_1d, axis, a, q, **kwargs)
        # apply_along_axis fills in collapsed axis with results.
        # Move that axis to the beginning to match percentile's
        # convention.
        if q.ndim != 0:
            result = np.moveaxis(result, axis, 0)

    if out is not None:
        out[...] = result
    return result


def _preprocess_quantile(a, q, axis=None, out=None, **kwargs):
    from astropy.utils.masked import Masked

    if isinstance(q, Masked) or out is not None and not isinstance(out, Masked):
        raise NotImplementedError

    a = Masked(a)
    q = np.asanyarray(q)
    if not _quantile_is_valid(q):
        raise ValueError("Quantiles must be in the range [0, 1]")

    if (interpolation := kwargs.pop("interpolation")) is not None:
        # we have to duplicate logic from np.quantile here to avoid
        # passing down the 'interpolation' keyword argument, as it's not
        # supported by np.lib._function_base_impl._quantile_unchecked
        kwargs["method"] = _check_interpolation_as_method(
            kwargs.get("method", "linear"), interpolation, "quantile"
        )
    return a, q, axis, out, kwargs


if NUMPY_LT_1_24:

    @dispatched_function
    def quantile(
        a,
        q,
        axis=None,
        out=None,
        overwrite_input=False,
        method="linear",
        keepdims=False,
        *,
        interpolation=None,
    ):
        a, q, axis, out, kwargs = _preprocess_quantile(
            a,
            q,
            axis,
            out,
            overwrite_input=overwrite_input,
            method=method,
            keepdims=keepdims,
            interpolation=interpolation,
        )
        keepdims = kwargs.pop("keepdims", False)
        r, k = _ureduce(a, func=_masked_quantile, q=q, axis=axis, out=out, **kwargs)
        result = (r.reshape(q.shape + k) if keepdims else r) if out is None else out
        return result, None, None

elif NUMPY_LT_2_0:

    @dispatched_function
    def quantile(
        a,
        q,
        axis=None,
        out=None,
        overwrite_input=False,
        method="linear",
        keepdims=False,
        *,
        interpolation=None,
    ):
        a, q, axis, out, kwargs = _preprocess_quantile(
            a,
            q,
            axis,
            out,
            overwrite_input=overwrite_input,
            method=method,
            keepdims=keepdims,
            interpolation=interpolation,
        )
        result = _ureduce(a, func=_masked_quantile, q=q, axis=axis, out=out, **kwargs)
        return result, None, None

else:

    @dispatched_function
    def quantile(
        a,
        q,
        axis=None,
        out=None,
        overwrite_input=False,
        method="linear",
        keepdims=False,
        *,
        weights=None,
        interpolation=None,
    ):
        a, q, axis, out, kwargs = _preprocess_quantile(
            a,
            q,
            axis,
            out,
            overwrite_input=overwrite_input,
            method=method,
            keepdims=keepdims,
            weights=weights,
            interpolation=interpolation,
        )
        result = _ureduce(a, func=_masked_quantile, q=q, axis=axis, out=out, **kwargs)
        return result, None, None


@dispatched_function
def percentile(a, q, *args, **kwargs):
    q = np.true_divide(q, 100)
    return quantile(a, q, *args, **kwargs)


@dispatched_function
def array_equal(a1, a2, equal_nan=False):
    (a1d, a2d), (a1m, a2m) = _get_data_and_masks(a1, a2)
    if a1d.shape != a2d.shape:
        return False, None, None

    equal = a1d == a2d
    if equal_nan:
        equal |= np.isnan(a1d) & np.isnan(a2d)
    return bool((equal | a1m | a2m).all()), None, None


@dispatched_function
def array_equiv(a1, a2):
    return bool((a1 == a2).all()), None, None


@dispatched_function
def where(condition, *args):
    from astropy.utils.masked import Masked

    if not args:
        return condition.nonzero(), None, None

    condition, c_mask = Masked._get_data_and_mask(condition)

    data, masks = _get_data_and_masks(*args)
    unmasked = np.where(condition, *data)
    mask = np.where(condition, *masks)
    if c_mask is not None:
        mask |= c_mask
    return Masked(unmasked, mask=mask), None, None


@dispatched_function
def choose(a, choices, out=None, mode="raise"):
    """Construct an array from an index array and a set of arrays to choose from.

    Like `numpy.choose`.  Masked indices in ``a`` will lead to masked output
    values and underlying data values are ignored if out of bounds (for
    ``mode='raise'``).  Any values masked in ``choices`` will be propagated
    if chosen.

    """
    from astropy.utils.masked import Masked

    a_data, a_mask = Masked._get_data_and_mask(a)
    if a_mask is not None and mode == "raise":
        # Avoid raising on masked indices.
        a_data = a.filled(fill_value=0)

    kwargs = {"mode": mode}
    if out is not None:
        if not isinstance(out, Masked):
            raise NotImplementedError
        kwargs["out"] = out.unmasked

    data, masks = _get_data_and_masks(*choices)
    data_chosen = np.choose(a_data, data, **kwargs)
    if out is not None:
        kwargs["out"] = out.mask

    mask_chosen = np.choose(a_data, masks, **kwargs)
    if a_mask is not None:
        mask_chosen |= a_mask

    return Masked(data_chosen, mask_chosen) if out is None else out, None, None


@apply_to_both
def select(condlist, choicelist, default=0):
    """Return an array drawn from elements in choicelist, depending on conditions.

    Like `numpy.select`, with masks in ``choicelist`` are propagated.
    Any masks in ``condlist`` are ignored.

    """
    from astropy.utils.masked import Masked

    condlist = [c.unmasked if isinstance(c, Masked) else c for c in condlist]

    data_list, mask_list = _get_data_and_masks(*choicelist)
    default = Masked(default) if default is not np.ma.masked else Masked(0, mask=True)
    return (
        (condlist, data_list, default.unmasked),
        (condlist, mask_list, default.mask),
        {},
        None,
    )


@dispatched_function
def piecewise(x, condlist, funclist, *args, **kw):
    """Evaluate a piecewise-defined function.

    Like `numpy.piecewise` but for masked input array ``x``.
    Any masks in ``condlist`` are ignored.

    """
    # Copied implementation from numpy.lib._function_base_impl.piecewise,
    # just to ensure output is Masked.
    n2 = len(funclist)
    # undocumented: single condition is promoted to a list of one condition
    if np.isscalar(condlist) or (
        not isinstance(condlist[0], (list, np.ndarray)) and x.ndim != 0
    ):  # pragma: no cover
        condlist = [condlist]

    condlist = np.array(condlist, dtype=bool)
    n = len(condlist)

    if n == n2 - 1:  # compute the "otherwise" condition.
        condelse = ~np.any(condlist, axis=0, keepdims=True)
        condlist = np.concatenate([condlist, condelse], axis=0)
        n += 1
    elif n != n2:
        raise ValueError(
            f"with {n} condition(s), either {n} or {n + 1} functions are expected"
        )

    # The one real change...
    y = np.zeros_like(x)
    where = []
    what = []
    for k in range(n):
        item = funclist[k]
        if not callable(item):
            where.append(condlist[k])
            what.append(item)
        else:
            vals = x[condlist[k]]
            if vals.size > 0:
                where.append(condlist[k])
                what.append(item(vals, *args, **kw))

    for item, value in zip(where, what):
        y[item] = value

    return y, None, None


@dispatched_function
def interp(x, xp, fp, *args, **kwargs):
    """One-dimensional linear interpolation.

    Like `numpy.interp`, but any masked points in ``xp`` and ``fp``
    are ignored.  Any masked values in ``x`` will still be evaluated,
    but masked on output.
    """
    from astropy.utils.masked import Masked

    xd, xm = Masked._get_data_and_mask(x)
    if isinstance(xp, Masked) or isinstance(fp, Masked):
        (xp, fp), (xpm, fpm) = _get_data_and_masks(xp, fp)
        if xp.ndim == fp.ndim == 1:
            # Avoid making arrays 1-D; will just raise below.
            m = xpm | fpm
            xp = xp[~m]
            fp = fp[~m]

    result = np.interp(xd, xp, fp, *args, **kwargs)
    return (result if xm is None else Masked(result, xm.copy())), None, None


@dispatched_function
def lexsort(keys, axis=-1):
    """Perform an indirect stable sort using a sequence of keys.

    Like `numpy.lexsort` but for possibly masked ``keys``.  Masked
    values are sorted towards the end for each key.
    """
    # Sort masks to the end.
    from .core import Masked

    new_keys = []
    for key in keys:
        if isinstance(key, Masked):
            # If there are other keys below, want to be sure that
            # for masked values, those other keys set the order.
            new_key = key.unmasked
            if new_keys and key.mask.any():
                new_key = new_key.copy()
                new_key[key.mask] = new_key.flat[0]
            new_keys.extend([new_key, key.mask])
        else:
            new_keys.append(key)

    return np.lexsort(new_keys, axis=axis), None, None


@dispatched_function
def apply_over_axes(func, a, axes):
    # Copied straight from numpy/lib/shape_base, just to omit its
    # val = asarray(a); if only it had been asanyarray, or just not there
    # since a is assumed to an an array in the next line...
    # Which is what we do here - we can only get here if it is Masked.
    val = a
    N = a.ndim
    if np.array(axes).ndim == 0:
        axes = (axes,)
    for axis in axes:
        if axis < 0:
            axis = N + axis
        args = (val, axis)
        res = func(*args)
        if res.ndim == val.ndim:
            val = res
        else:
            res = np.expand_dims(res, axis)
            if res.ndim == val.ndim:
                val = res
            else:
                raise ValueError(
                    "function is not returning an array of the correct shape"
                )

    return val, None, None


class MaskedFormat:
    """Formatter for masked array scalars.

    For use in `numpy.array2string`, wrapping the regular formatters such
    that if a value is masked, its formatted string is replaced.

    Typically initialized using the ``from_data`` class method.
    """

    def __init__(self, format_function):
        self.format_function = format_function
        # Special case for structured void and subarray: we need to make all the
        # format functions for the items masked as well.
        # TODO: maybe is a separate class is more logical?
        ffs = getattr(format_function, "format_functions", None)
        if ffs:
            # StructuredVoidFormat: multiple format functions to be changed.
            self.format_function.format_functions = [MaskedFormat(ff) for ff in ffs]

        ff = getattr(format_function, "format_function", None)
        if ff:
            # SubarrayFormat: change format function for the elements.
            self.format_function.format_function = MaskedFormat(ff)

    def __call__(self, x):
        if x.dtype.names:
            # The replacement of x with a list is needed because the function
            # inside StructuredVoidFormat iterates over x, which works for an
            # np.void but not an array scalar.
            return self.format_function([x[field] for field in x.dtype.names])

        if x.shape:
            # For a subarray pass on the data directly, since the
            # items will be iterated on inside the function.
            return self.format_function(x)

        # Single element: first just typeset it normally, replace with masked
        # string if needed.
        string = self.format_function(x.unmasked[()])
        if x.mask:
            # Strikethrough would be neat, but terminal needs a different
            # formatting than, say, jupyter notebook.
            # return "\x1B[9m"+string+"\x1B[29m"
            # return ''.join(s+'\u0336' for s in string)
            n = min(3, max(1, len(string)))
            return " " * (len(string) - n) + "\u2014" * n
        else:
            return string

    @classmethod
    def from_data(cls, data, **options):
        if NUMPY_LT_2_0:
            from numpy.core.arrayprint import _get_format_function
        else:
            from numpy._core.arrayprint import _get_format_function

        return cls(_get_format_function(data, **options))


def _array2string(a, options, separator=" ", prefix=""):
    # Mostly copied from numpy.core.arrayprint, except:
    # - The format function is wrapped in a mask-aware class;
    # - Arrays scalars are not cast as arrays.
    if NUMPY_LT_2_0:
        from numpy.core.arrayprint import _formatArray, _leading_trailing
    else:
        from numpy._core.arrayprint import _formatArray, _leading_trailing

    data = np.asarray(a)

    if a.size > options["threshold"]:
        summary_insert = "..."
        data = _leading_trailing(data, options["edgeitems"])
    else:
        summary_insert = ""

    # find the right formatting function for the array
    format_function = MaskedFormat.from_data(data, **options)

    # skip over "["
    next_line_prefix = " "
    # skip over array(
    next_line_prefix += " " * len(prefix)

    lst = _formatArray(
        a,
        format_function,
        options["linewidth"],
        next_line_prefix,
        separator,
        options["edgeitems"],
        summary_insert,
        options["legacy"],
    )
    return lst


@dispatched_function
def array2string(
    a,
    max_line_width=None,
    precision=None,
    suppress_small=None,
    separator=" ",
    prefix="",
    style=np._NoValue,
    formatter=None,
    threshold=None,
    edgeitems=None,
    sign=None,
    floatmode=None,
    suffix="",
    *,
    legacy=None,
):
    # Copied from numpy.core.arrayprint, but using _array2string above.
    if NUMPY_LT_2_1:
        if NUMPY_LT_2_0:
            from numpy.core.arrayprint import _format_options
        else:
            from numpy._core.arrayprint import _format_options
        options = _format_options.copy()
    else:
        from numpy._core.printoptions import format_options

        options = format_options.get().copy()

    if NUMPY_LT_2_0:
        from numpy.core.arrayprint import _make_options_dict
    else:
        from numpy._core.arrayprint import _make_options_dict

    overrides = _make_options_dict(
        precision,
        threshold,
        edgeitems,
        max_line_width,
        suppress_small,
        None,
        None,
        sign,
        formatter,
        floatmode,
    )
    options.update(overrides)

    options["linewidth"] -= len(suffix)

    if legacy is not None:
        warnings.warn(f"{legacy=} is ignored.")

    # treat as a null array if any of shape elements == 0
    if a.size == 0:
        result = "[]"
    else:
        result = _array2string(a, options, separator, prefix)

    return result, None, None


def _array_str_scalar(x):
    # This wraps np.array_str for use as a format function in
    # MaskedFormat. We cannot use it directly as format functions
    # expect numpy scalars, while np.array_str expects an array.
    return np.array_str(np.array(x))


@dispatched_function
def array_str(a, max_line_width=None, precision=None, suppress_small=None):
    # Override to change special treatment of array scalars, since the numpy
    # code turns the masked array scalar into a regular array scalar.
    # By going through MaskedFormat, we can replace the string as needed.
    if a.shape == ():
        if a.dtype.names is None:
            return MaskedFormat(_array_str_scalar)(a), None, None
        elif not NUMPY_LT_2_0:
            from numpy._core.arrayprint import StructuredVoidFormat

            # Following numpy._core.arrayprint._void_scalar_to_string
            if NUMPY_LT_2_1:
                from numpy._core.arrayprint import _format_options

                options = _format_options.copy()
            else:
                from numpy._core.printoptions import format_options

                options = format_options.get().copy()

            if options.get("formatter") is None:
                options["formatter"] = {}
            options["formatter"].setdefault("float_kind", str)
            return (
                MaskedFormat(StructuredVoidFormat.from_data(a, **options))(a),
                None,
                None,
            )

    return array2string(a, max_line_width, precision, suppress_small, " ", "")


# For the nanfunctions, we just treat any nan as an additional mask.
_nanfunc_fill_values = {"nansum": 0, "nancumsum": 0, "nanprod": 1, "nancumprod": 1}


def masked_nanfunc(nanfuncname):
    np_func = getattr(np, nanfuncname[3:])
    fill_value = _nanfunc_fill_values.get(nanfuncname)

    def nanfunc(a, *args, **kwargs):
        from astropy.utils.masked import Masked

        a, mask = Masked._get_data_and_mask(a)
        if issubclass(a.dtype.type, np.inexact):
            nans = np.isnan(a)
            mask = nans if mask is None else (nans | mask)

        if mask is not None:
            a = Masked(a, mask)
            if fill_value is not None:
                a = a.filled(fill_value)

        return np_func(a, *args, **kwargs), None, None

    doc = f"Like `numpy.{nanfuncname}`, skipping masked values as well.\n\n"
    if fill_value is not None:
        # sum, cumsum, prod, cumprod
        doc += (
            f"Masked/NaN values are replaced with {fill_value}. "
            "The output is not masked."
        )
    elif "arg" in nanfuncname:
        doc += (
            "No exceptions are raised for fully masked/NaN slices.\n"
            "Instead, these give index 0."
        )
    else:
        doc += (
            "No warnings are given for fully masked/NaN slices.\n"
            "Instead, they are masked in the output."
        )

    nanfunc.__doc__ = doc
    nanfunc.__name__ = nanfuncname

    return nanfunc


_nplibnanfunctions = np.lib.nanfunctions if NUMPY_LT_2_0 else np.lib._nanfunctions_impl
for nanfuncname in _nplibnanfunctions.__all__:
    globals()[nanfuncname] = dispatched_function(
        masked_nanfunc(nanfuncname), helps=getattr(np, nanfuncname)
    )


@dispatched_function
def ediff1d(ary, to_end=None, to_begin=None):
    from astropy.utils.masked import Masked

    # ediff1d works fine if ary is Masked, but not if it is not (and
    # we got here because to_end and/or to_begin are Masked).
    if not isinstance(ary, Masked):
        ary = Masked(ary)

    return np.ediff1d.__wrapped__(ary, to_end, to_begin), None, None


def _in1d(ar1, ar2, assume_unique=False, invert=False, *, kind=None):
    # Copy sorting implementation from _arraysetops_impl._in1d;
    # the others cannot work in the presence of a mask.
    if kind == "table":
        raise ValueError(
            "The 'table' method is not supported for Masked arrays."
            "Please select 'sort' or None for kind."
        )

    # Straight copy from here on.
    if not assume_unique:
        ar1, rev_idx = np.unique(ar1, return_inverse=True)
        ar2 = np.unique(ar2)

    ar = np.concatenate((ar1, ar2))
    # We need this to be a stable sort, so always use 'mergesort'
    # here. The values from the first array should always come before
    # the values from the second array.
    order = ar.argsort(kind="mergesort")
    sar = ar[order]
    if invert:
        bool_ar = sar[1:] != sar[:-1]
    else:
        bool_ar = sar[1:] == sar[:-1]
    flag = np.concatenate((bool_ar, [invert]))
    ret = np.empty(ar.shape, dtype=bool)
    ret[order] = flag
    return ret[: len(ar1)] if assume_unique else ret[rev_idx]


def _copy_of_mask(a):
    mask = getattr(a, "mask", None)
    return mask.copy() if mask is not None else False


if NUMPY_LT_1_24:  # "kind" argument introduced in 1.24.

    @dispatched_function
    def in1d(ar1, ar2, assume_unique=False, invert=False):
        mask = _copy_of_mask(ar1).ravel()
        return _in1d(ar1, ar2, assume_unique, invert), mask, None

    @dispatched_function
    def isin(element, test_elements, assume_unique=False, invert=False):
        element = np.asanyarray(element)
        result = _in1d(element, test_elements, assume_unique, invert)
        result.shape = element.shape
        return result, _copy_of_mask(element), None

else:

    @dispatched_function
    def in1d(ar1, ar2, assume_unique=False, invert=False, *, kind=None):
        mask = _copy_of_mask(ar1).ravel()
        return _in1d(ar1, ar2, assume_unique, invert, kind=kind), mask, None

    @dispatched_function
    def isin(element, test_elements, assume_unique=False, invert=False, *, kind=None):
        element = np.asanyarray(element)
        result = _in1d(element, test_elements, assume_unique, invert, kind=kind)
        result.shape = element.shape
        return result, _copy_of_mask(element), None


@dispatched_function
def setdiff1d(ar1, ar2, assume_unique=False):
    # Again, mostly just to avoid an asarray call.
    if assume_unique:
        ar1 = np.asanyarray(ar1).ravel()
    else:
        ar1 = np.unique(ar1)
        ar2 = np.unique(ar2)
    return ar1[np.isin(ar1, ar2, assume_unique=True, invert=True)], None, None


# Add any dispatched or helper function that has a docstring to
# __all__, so they will be typeset by sphinx. The logic is that for
# those presumably the use of the mask is not entirely obvious.
__all__ += sorted(  # noqa: PLE0605
    helper.__name__
    for helper in (
        set(APPLY_TO_BOTH_FUNCTIONS.values()) | set(DISPATCHED_FUNCTIONS.values())
    )
    if helper.__doc__
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import re
import warnings
from pathlib import Path

import numpy as np
import pytest
from astropy_iers_data import (
    IERS_A_README,
    IERS_B_FILE,
    IERS_B_README,
    IERS_LEAP_SECOND_FILE,
)

from astropy import units as u
from astropy.config import set_temp_cache
from astropy.table import QTable
from astropy.tests.helper import CI, assert_quantity_allclose
from astropy.time import Time, TimeDelta
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyDeprecationWarning
from astropy.utils.iers import iers

FILE_NOT_FOUND_ERROR = getattr(__builtins__, "FileNotFoundError", OSError)

try:
    iers.IERS_A.open("finals2000A.all")  # check if IERS_A is available
except OSError:
    HAS_IERS_A = False
else:
    HAS_IERS_A = True

IERS_A_EXCERPT = get_pkg_data_filename(os.path.join("data", "iers_a_excerpt"))


def setup_module():
    # Need auto_download so that IERS_B won't be loaded and cause tests to
    # fail. Files to be downloaded are handled appropriately in the tests.
    iers.conf.auto_download = True


def teardown_module():
    # This setting is to be consistent with astropy/conftest.py
    iers.conf.auto_download = False


class TestBasic:
    """Basic tests that IERS_B returns correct values"""

    @pytest.mark.parametrize("iers_cls", (iers.IERS_B, iers.IERS))
    def test_simple(self, iers_cls):
        """Test the default behaviour for IERS_B and IERS."""
        # Arguably, IERS itself should not be used at all, but it used to
        # provide IERS_B by default so we check that it continues to do so.
        # Eventually, IERS should probably be deprecated.
        iers_cls.close()
        assert iers_cls.iers_table is None
        iers_tab = iers_cls.open()
        assert iers_cls.iers_table is not None
        assert iers_cls.iers_table is iers_tab
        assert isinstance(iers_tab, QTable)
        assert isinstance(iers_tab, iers.IERS_B)
        assert (iers_tab["UT1_UTC"].unit / u.second).is_unity()
        assert (iers_tab["PM_x"].unit / u.arcsecond).is_unity()
        assert (iers_tab["PM_y"].unit / u.arcsecond).is_unity()
        jd1 = np.array([2456108.5, 2456108.5, 2456108.5, 2456109.5, 2456109.5])
        jd2 = np.array([0.49999421, 0.99997685, 0.99998843, 0.0, 0.5])
        ut1_utc = iers_tab.ut1_utc(jd1, jd2)
        assert isinstance(ut1_utc, u.Quantity)
        assert (ut1_utc.unit / u.second).is_unity()
        # IERS files change at the 0.1 ms level; see gh-6981
        assert_quantity_allclose(
            ut1_utc,
            [-0.5868211, -0.5868184, -0.5868184, 0.4131816, 0.41328895] * u.s,
            atol=0.1 * u.ms,
        )
        # should be future-proof; surely we've moved to another planet by then
        with pytest.raises(IndexError):
            ut1_utc2, status2 = iers_tab.ut1_utc(1e11, 0.0)
        # also check it returns the right status
        ut1_utc2, status2 = iers_tab.ut1_utc(jd1, jd2, return_status=True)
        assert np.all(status2 == iers.FROM_IERS_B)
        ut1_utc4, status4 = iers_tab.ut1_utc(1e11, 0.0, return_status=True)
        assert status4 == iers.TIME_BEYOND_IERS_RANGE

        # check it works via Time too
        t = Time(jd1, jd2, format="jd", scale="utc")
        ut1_utc3 = iers_tab.ut1_utc(t)
        assert_quantity_allclose(
            ut1_utc3,
            [-0.5868211, -0.5868184, -0.5868184, 0.4131816, 0.41328895] * u.s,
            atol=0.1 * u.ms,
        )

        # Table behaves properly as a table (e.g. can be sliced)
        assert len(iers_tab[:2]) == 2

    def test_open_filename(self):
        iers.IERS_B.close()
        iers.IERS_B.open(iers.IERS_B_FILE)
        assert iers.IERS_B.iers_table is not None
        assert isinstance(iers.IERS_B.iers_table, QTable)
        iers.IERS_B.close()
        with pytest.raises(FILE_NOT_FOUND_ERROR):
            iers.IERS_B.open("surely this does not exist")

    def test_open_network_url(self):
        iers.IERS_A.close()
        iers.IERS_A.open(Path(IERS_A_EXCERPT).as_uri())
        assert iers.IERS_A.iers_table is not None
        assert isinstance(iers.IERS_A.iers_table, QTable)
        iers.IERS_A.close()


def test_IERS_B_old_style_excerpt():
    """Check that the instructions given in `IERS_B.read` actually work."""
    # If this test is changed, be sure to also adjust the instructions.
    #
    # TODO: this test and the note can probably be removed after
    # enough time has passed that old-style IERS_B files are simply
    # not around any more, say in 2025.  If so, also remove the excerpt
    # and the ReadMe.eopc04_IAU2000 file.
    old_style_file = get_pkg_data_filename(
        os.path.join("data", "iers_b_old_style_excerpt")
    )
    excerpt = iers.IERS_B.read(
        old_style_file,
        readme=get_pkg_data_filename(
            "data/ReadMe.eopc04_IAU2000", package="astropy.utils.iers"
        ),
        data_start=14,
    )
    assert isinstance(excerpt, QTable)
    assert "PM_x_dot" not in excerpt.colnames


class TestIERS_AExcerpt:
    @classmethod
    def teardown_class(self):
        iers.IERS_A.close()

    def test_simple(self):
        # Test the IERS A reader. It is also a regression tests that ensures
        # values do not get overridden by IERS B; see #4933.
        iers_tab = iers.IERS_A.open(IERS_A_EXCERPT)

        assert (iers_tab["UT1_UTC"].unit / u.second).is_unity()
        assert "P" in iers_tab["UT1Flag"]
        assert "I" in iers_tab["UT1Flag"]
        assert "B" in iers_tab["UT1Flag"]
        assert np.all(
            (iers_tab["UT1Flag"] == "I")
            | (iers_tab["UT1Flag"] == "P")
            | (iers_tab["UT1Flag"] == "B")
        )

        assert (iers_tab["dX_2000A"].unit / u.marcsec).is_unity()
        assert (iers_tab["dY_2000A"].unit / u.marcsec).is_unity()
        assert "P" in iers_tab["NutFlag"]
        assert "I" in iers_tab["NutFlag"]
        assert "B" in iers_tab["NutFlag"]
        assert np.all(
            (iers_tab["NutFlag"] == "P")
            | (iers_tab["NutFlag"] == "I")
            | (iers_tab["NutFlag"] == "B")
        )

        assert (iers_tab["PM_x"].unit / u.arcsecond).is_unity()
        assert (iers_tab["PM_y"].unit / u.arcsecond).is_unity()
        assert "P" in iers_tab["PolPMFlag"]
        assert "I" in iers_tab["PolPMFlag"]
        assert "B" in iers_tab["PolPMFlag"]
        assert np.all(
            (iers_tab["PolPMFlag"] == "P")
            | (iers_tab["PolPMFlag"] == "I")
            | (iers_tab["PolPMFlag"] == "B")
        )

        t = Time([57053.0, 57054.0, 57055.0], format="mjd")
        ut1_utc, status = iers_tab.ut1_utc(t, return_status=True)
        assert status[0] == iers.FROM_IERS_B
        assert np.all(status[1:] == iers.FROM_IERS_A)
        # These values are *exactly* as given in the table, so they should
        # match to double precision accuracy.
        assert_quantity_allclose(
            ut1_utc, [-0.4916557, -0.4925323, -0.4934373] * u.s, atol=0.1 * u.ms
        )

        dcip_x, dcip_y, status = iers_tab.dcip_xy(t, return_status=True)
        assert status[0] == iers.FROM_IERS_B
        assert np.all(status[1:] == iers.FROM_IERS_A)
        # These values are *exactly* as given in the table, so they should
        # match to double precision accuracy.
        print(dcip_x)
        print(dcip_y)
        assert_quantity_allclose(
            dcip_x, [-0.086, -0.093, -0.087] * u.marcsec, atol=1.0 * u.narcsec
        )
        assert_quantity_allclose(
            dcip_y, [0.094, 0.081, 0.072] * u.marcsec, atol=1 * u.narcsec
        )

        pm_x, pm_y, status = iers_tab.pm_xy(t, return_status=True)
        assert status[0] == iers.FROM_IERS_B
        assert np.all(status[1:] == iers.FROM_IERS_A)
        assert_quantity_allclose(
            pm_x, [0.003734, 0.004581, 0.004623] * u.arcsec, atol=0.1 * u.marcsec
        )
        assert_quantity_allclose(
            pm_y, [0.310824, 0.313150, 0.315517] * u.arcsec, atol=0.1 * u.marcsec
        )

        # Table behaves properly as a table (e.g. can be sliced)
        assert len(iers_tab[:2]) == 2


@pytest.mark.skipif(not HAS_IERS_A, reason="requires IERS_A")
class TestIERS_A:
    @classmethod
    def teardown_class(self):
        iers.IERS_A.close()

    def test_simple(self):
        """Test that open() by default reads a 'finals2000A.all' file."""
        # Ensure we remove any cached table (gh-5131).
        iers.IERS_A.close()
        iers_tab = iers.IERS_A.open()
        jd1 = np.array([2456108.5, 2456108.5, 2456108.5, 2456109.5, 2456109.5])
        jd2 = np.array([0.49999421, 0.99997685, 0.99998843, 0.0, 0.5])
        ut1_utc, status = iers_tab.ut1_utc(jd1, jd2, return_status=True)
        assert np.all(status == iers.FROM_IERS_B)
        assert_quantity_allclose(
            ut1_utc,
            [-0.5868211, -0.5868184, -0.5868184, 0.4131816, 0.41328895] * u.s,
            atol=0.1 * u.ms,
        )
        ut1_utc2, status2 = iers_tab.ut1_utc(1e11, 0.0, return_status=True)
        assert status2 == iers.TIME_BEYOND_IERS_RANGE

        tnow = Time.now()

        ut1_utc3, status3 = iers_tab.ut1_utc(tnow, return_status=True)
        assert status3 == iers.FROM_IERS_A_PREDICTION
        assert ut1_utc3 != 0.0


class TestIERS_Auto:
    def setup_class(self):
        """Set up useful data for the tests."""
        self.N = 40
        self.ame = 30.0
        self.iers_a_file_1 = get_pkg_data_filename(
            os.path.join("data", "finals2000A-2016-02-30-test")
        )
        self.iers_a_file_2 = get_pkg_data_filename(
            os.path.join("data", "finals2000A-2016-04-30-test")
        )
        self.iers_a_url_1 = Path(self.iers_a_file_1).as_uri()
        self.iers_a_url_2 = Path(self.iers_a_file_2).as_uri()
        self.t = Time.now() + TimeDelta(10, format="jd") * np.arange(self.N)

    def teardown_method(self, method):
        """Run this after every test."""
        iers.IERS_Auto.close()

    def test_interpolate_error_formatting(self):
        """Regression test: make sure the error message in
        IERS_Auto._check_interpolate_indices() is formatted correctly.
        """
        with iers.conf.set_temp("iers_auto_url", self.iers_a_url_1):
            with iers.conf.set_temp("iers_auto_url_mirror", self.iers_a_url_1):
                with iers.conf.set_temp("auto_max_age", self.ame):
                    with pytest.raises(
                        ValueError,
                        match=re.escape(iers.INTERPOLATE_ERROR.format(self.ame)),
                    ):
                        iers_table = iers.IERS_Auto.open()
                        with warnings.catch_warnings():
                            # Ignoring this if it comes up -- IERS_Auto predictive
                            # values are older than 30.0 days but downloading the
                            # latest table did not find newer values
                            warnings.simplefilter("ignore", iers.IERSStaleWarning)
                            iers_table.ut1_utc(self.t.jd1, self.t.jd2)

    def test_auto_max_age_none(self):
        """Make sure that iers.INTERPOLATE_ERROR's advice about setting
        auto_max_age = None actually works.
        """
        with iers.conf.set_temp("iers_auto_url", self.iers_a_url_1):
            with iers.conf.set_temp("auto_max_age", None):
                iers_table = iers.IERS_Auto.open()
                delta = iers_table.ut1_utc(self.t.jd1, self.t.jd2)
        assert isinstance(delta, np.ndarray)
        assert delta.shape == (self.N,)
        assert_quantity_allclose(delta, np.array([-0.2246227] * self.N) * u.s)

    def test_auto_max_age_minimum(self):
        """Check that the minimum auto_max_age is enforced."""
        with iers.conf.set_temp("iers_auto_url", self.iers_a_url_1):
            with iers.conf.set_temp("auto_max_age", 5.0):
                with pytest.raises(
                    ValueError,
                    match=(
                        r"IERS auto_max_age configuration value must be larger than 10"
                        r" days"
                    ),
                ):
                    iers_table = iers.IERS_Auto.open()
                    _ = iers_table.ut1_utc(self.t.jd1, self.t.jd2)

    def test_no_auto_download(self):
        with iers.conf.set_temp("auto_download", False):
            t = iers.IERS_Auto.open()
        assert type(t) is iers.IERS_B

    @pytest.mark.remote_data
    def test_simple(self):
        with iers.conf.set_temp("iers_auto_url", self.iers_a_url_1):
            dat = iers.IERS_Auto.open()
            assert dat["MJD"][0] == 57359.0 * u.d
            assert dat["MJD"][-1] == 57539.0 * u.d

            # Pretend we are accessing at a time 7 days after start of predictive data
            predictive_mjd = dat.meta["predictive_mjd"]
            dat._time_now = Time(predictive_mjd, format="mjd") + 7 * u.d

            # Look at times before and after the test file begins.  0.1292934 is
            # the IERS-B value from MJD=57359.  The value in
            # finals2000A-2016-02-30-test has been replaced at this point.
            assert np.allclose(
                dat.ut1_utc(Time(50000, format="mjd").jd).value, 0.1292934
            )
            assert np.allclose(
                dat.ut1_utc(Time(60000, format="mjd").jd).value, -0.2246227
            )

            # Now pretend we are accessing at time 60 days after start of predictive data.
            # There will be a warning when downloading the file doesn't give new data
            # and an exception when extrapolating into the future with insufficient data.
            dat._time_now = Time(predictive_mjd, format="mjd") + 60 * u.d
            assert np.allclose(
                dat.ut1_utc(Time(50000, format="mjd").jd).value, 0.1292934
            )
            with (
                pytest.warns(
                    iers.IERSStaleWarning, match="IERS_Auto predictive values are older"
                ) as warns,
                pytest.raises(
                    ValueError,
                    match="interpolating from IERS_Auto using predictive values",
                ),
            ):
                dat.ut1_utc(Time(60000, format="mjd").jd)
            assert len(warns) == 1

            # Warning only if we are getting return status
            with pytest.warns(
                iers.IERSStaleWarning, match="IERS_Auto predictive values are older"
            ) as warns:
                dat.ut1_utc(Time(60000, format="mjd").jd, return_status=True)
            assert len(warns) == 1

            # Now set auto_max_age = None which says that we don't care how old the
            # available IERS-A file is.  There should be no warnings or exceptions.
            with iers.conf.set_temp("auto_max_age", None):
                dat.ut1_utc(Time(60000, format="mjd").jd)

        # Now point to a later file with same values but MJD increased by
        # 60 days and see that things work.  dat._time_now is still the same value
        # as before, i.e. right around the start of predictive values for the new file.
        # (In other words this is like downloading the latest file online right now).
        with iers.conf.set_temp("iers_auto_url", self.iers_a_url_2):
            # Look at times before and after the test file begins.  This forces a new download.
            assert np.allclose(
                dat.ut1_utc(Time(50000, format="mjd").jd).value, 0.1292934
            )
            assert np.allclose(dat.ut1_utc(Time(60000, format="mjd").jd).value, -0.3)

            # Now the time range should be different.
            assert dat["MJD"][0] == 57359.0 * u.d
            assert dat["MJD"][-1] == (57539.0 + 60) * u.d


@pytest.mark.parametrize("query", ["ut1_utc", "pm_xy"])
@pytest.mark.parametrize("jd", [np.array([]), Time([], format="mjd")])
@pytest.mark.parametrize("return_status", [False, True])
def test_empty_mjd(query, jd, return_status):
    # Regression test for gh-17008
    with iers.conf.set_temp("auto_download", False):
        iers_table = iers.IERS_Auto.open()
        result = getattr(iers_table, query)(jd, return_status=return_status)
        n_exp = (1 if query == "ut1_utc" else 2) + (1 if return_status else 0)
        if n_exp == 1:
            assert isinstance(result, np.ndarray)
            assert result.size == 0
        else:
            assert len(result) == n_exp
            assert all(r.size == 0 for r in result)


@pytest.mark.remote_data
def test_IERS_B_parameters_loading_into_IERS_Auto():
    A = iers.IERS_Auto.open()
    B = iers.IERS_B.open()

    ok_A = A["MJD"] <= B["MJD"][-1]
    assert not np.all(ok_A), "IERS B covers all of IERS A: should not happen"

    # We only overwrite IERS_B values in the IERS_A table that were already
    # there in the first place.  Better take that into account.
    ok_A &= np.isfinite(A["UT1_UTC_B"])

    i_B = np.searchsorted(B["MJD"], A["MJD"][ok_A])

    assert np.all(np.diff(i_B) == 1), "Valid region not contiguous"
    assert np.all(A["MJD"][ok_A] == B["MJD"][i_B])
    # Check that values are copied correctly.  Since units are not
    # necessarily the same, we use allclose with very strict tolerance.
    for name in ("UT1_UTC", "PM_x", "PM_y", "dX_2000A", "dY_2000A"):
        assert_quantity_allclose(
            A[name][ok_A],
            B[name][i_B],
            rtol=1e-15,
            err_msg=(
                f"Bug #9206 IERS B parameter {name} not copied over "
                "correctly to IERS Auto"
            ),
        )


# Issue with FTP, rework test into previous one when it's fixed
@pytest.mark.skipif(CI, reason="Flaky on CI")
@pytest.mark.remote_data
def test_iers_a_dl():
    iersa_tab = iers.IERS_A.open(iers.IERS_A_URL, cache=False)
    try:
        # some basic checks to ensure the format makes sense
        assert len(iersa_tab) > 0
        assert "UT1_UTC_A" in iersa_tab.colnames
    finally:
        iers.IERS_A.close()


@pytest.mark.remote_data
def test_iers_a_dl_mirror():
    iersa_tab = iers.IERS_A.open(iers.IERS_A_URL_MIRROR, cache=False)
    try:
        # some basic checks to ensure the format makes sense
        assert len(iersa_tab) > 0
        assert "UT1_UTC_A" in iersa_tab.colnames
    finally:
        iers.IERS_A.close()


@pytest.mark.remote_data
def test_iers_b_dl():
    iersb_tab = iers.IERS_B.open(iers.IERS_B_URL, cache=False)
    try:
        # some basic checks to ensure the format makes sense
        assert len(iersb_tab) > 0
        assert "UT1_UTC" in iersb_tab.colnames
    finally:
        iers.IERS_B.close()


@pytest.mark.remote_data
def test_iers_out_of_range_handling(tmp_path):
    # Make sure we don't have IERS-A data available anywhere
    with set_temp_cache(tmp_path):
        iers.IERS_A.close()
        iers.IERS_Auto.close()
        iers.IERS.close()
        now = Time.now()
        with iers.conf.set_temp("auto_download", False):
            # Should be fine with built-in IERS_B
            (now - 300 * u.day).ut1

            # Default is to raise an error
            match = r"\(some\) times are outside of range covered by IERS table"
            with pytest.raises(iers.IERSRangeError, match=match):
                (now + 100 * u.day).ut1

            with iers.conf.set_temp("iers_degraded_accuracy", "warn"):
                with pytest.warns(iers.IERSDegradedAccuracyWarning, match=match):
                    (now + 100 * u.day).ut1

            with iers.conf.set_temp("iers_degraded_accuracy", "ignore"):
                (now + 100 * u.day).ut1


@pytest.mark.remote_data
def test_iers_download_error_handling(tmp_path):
    # Make sure we don't have IERS-A data available anywhere
    with set_temp_cache(tmp_path):
        iers.IERS_A.close()
        iers.IERS_Auto.close()
        iers.IERS.close()
        now = Time.now()

        # bad site name
        with iers.conf.set_temp("iers_auto_url", "FAIL FAIL"):
            # site that exists but doesn't have IERS data
            with iers.conf.set_temp("iers_auto_url_mirror", "https://google.com"):
                with pytest.warns(iers.IERSWarning) as record:
                    with iers.conf.set_temp("iers_degraded_accuracy", "ignore"):
                        (now + 100 * u.day).ut1

                assert len(record) == 3
                assert str(record[0].message).startswith(
                    "failed to download FAIL FAIL: Malformed URL"
                )
                assert str(record[1].message).startswith(
                    "malformed IERS table from https://google.com"
                )
                assert str(record[2].message).startswith(
                    "unable to download valid IERS file, using local IERS-B"
                )


OLD_DATA_FILES = {
    "Leap_Second.dat": IERS_LEAP_SECOND_FILE,
    "ReadMe.finals2000A": IERS_A_README,
    "ReadMe.eopc04": IERS_B_README,
    "eopc04.1962-now": IERS_B_FILE,
}


@pytest.mark.parametrize("data_file", sorted(OLD_DATA_FILES))
def test_get_pkg_data_filename_backcompat(data_file):
    # Check that get_pkg_data_filename continues to work without breakage
    # if users use it to access IERS tables and READMEs that used to be in
    # astropy/utils/iers/data.

    with pytest.warns(
        AstropyDeprecationWarning,
        match=f"Accessing {data_file} in this way is deprecated",
    ):
        filename = get_pkg_data_filename(
            "data/" + data_file, package="astropy.utils.iers"
        )

    assert filename == OLD_DATA_FILES[data_file]
